import operator

from igraph._igraph import (
    EdgeSeq as _EdgeSeq,
    VertexSeq as _VertexSeq,
)


class VertexSeq(_VertexSeq):
    """Class representing a sequence of vertices in the graph.

    This class is most easily accessed by the C{vs} field of the
    L{Graph} object, which returns an ordered sequence of all vertices in
    the graph. The vertex sequence can be refined by invoking the
    L{VertexSeq.select()} method. L{VertexSeq.select()} can also be
    accessed by simply calling the L{VertexSeq} object.

    An alternative way to create a vertex sequence referring to a given
    graph is to use the constructor directly:

      >>> g = Graph.Full(3)
      >>> vs = VertexSeq(g)
      >>> restricted_vs = VertexSeq(g, [0, 1])

    The individual vertices can be accessed by indexing the vertex sequence
    object. It can be used as an iterable as well, or even in a list
    comprehension:

      >>> g=Graph.Full(3)
      >>> for v in g.vs:
      ...   v["value"] = v.index ** 2
      ...
      >>> [v["value"] ** 0.5 for v in g.vs]
      [0.0, 1.0, 2.0]

    The vertex set can also be used as a dictionary where the keys are the
    attribute names. The values corresponding to the keys are the values
    of the given attribute for every vertex selected by the sequence.

      >>> g=Graph.Full(3)
      >>> for idx, v in enumerate(g.vs):
      ...   v["weight"] = idx*(idx+1)
      ...
      >>> g.vs["weight"]
      [0, 2, 6]
      >>> g.vs.select(1,2)["weight"] = [10, 20]
      >>> g.vs["weight"]
      [0, 10, 20]

    If you specify a sequence that is shorter than the number of vertices in
    the VertexSeq, the sequence is reused:

      >>> g = Graph.Tree(7, 2)
      >>> g.vs["color"] = ["red", "green"]
      >>> g.vs["color"]
      ['red', 'green', 'red', 'green', 'red', 'green', 'red']

    You can even pass a single string or integer, it will be considered as a
    sequence of length 1:

      >>> g.vs["color"] = "red"
      >>> g.vs["color"]
      ['red', 'red', 'red', 'red', 'red', 'red', 'red']

    Some methods of the vertex sequences are simply proxy methods to the
    corresponding methods in the L{Graph} object. One such example is
    C{VertexSeq.degree()}:

      >>> g=Graph.Tree(7, 2)
      >>> g.vs.degree()
      [2, 3, 3, 1, 1, 1, 1]
      >>> g.vs.degree() == g.degree()
      True
    """

    def attributes(self):
        """Returns the list of all the vertex attributes in the graph
        associated to this vertex sequence."""
        return self.graph.vertex_attributes()

    def find(self, *args, **kwds):
        """Returns the first vertex of the vertex sequence that matches some
        criteria.

        The selection criteria are equal to the ones allowed by L{VertexSeq.select}.
        See L{VertexSeq.select} for more details.

        For instance, to find the first vertex with name C{foo} in graph C{g}:

            >>> g.vs.find(name="foo")            #doctest:+SKIP

        To find an arbitrary isolated vertex:

            >>> g.vs.find(_degree=0)             #doctest:+SKIP
        """
        # Shortcut: if "name" is in kwds, there are no positional arguments,
        # and the specified name is a string, we try that first because that
        # attribute is indexed. Note that we cannot do this if name is an
        # integer, because it would then translate to g.vs.select(name), which
        # searches by _index_ if the argument is an integer
        if not args:
            if "name" in kwds:
                name = kwds.pop("name")
            elif "name_eq" in kwds:
                name = kwds.pop("name_eq")
            else:
                name = None

            if name is not None:
                if isinstance(name, str):
                    args = [name]
                else:
                    # put back what we popped
                    kwds["name"] = name

        if args:
            # Selecting first based on positional arguments, then checking
            # the criteria specified by the (remaining) keyword arguments
            vertex = _VertexSeq.find(self, *args)
            if not kwds:
                return vertex
            vs = self.graph.vs.select(vertex.index)
        else:
            vs = self

        # Selecting based on keyword arguments
        vs = vs.select(**kwds)
        if vs:
            return vs[0]
        raise ValueError("no such vertex")

    def select(self, *args, **kwds):
        """Selects a subset of the vertex sequence based on some criteria

        The selection criteria can be specified by the positional and the keyword
        arguments. Positional arguments are always processed before keyword
        arguments.

          - If the first positional argument is C{None}, an empty sequence is
            returned.

          - If the first positional argument is a callable object, the object
            will be called for every vertex in the sequence. If it returns
            C{True}, the vertex will be included, otherwise it will
            be excluded.

          - If the first positional argument is an iterable, it must return
            integers and they will be considered as indices of the current
            vertex set (NOT the whole vertex set of the graph -- the
            difference matters when one filters a vertex set that has
            already been filtered by a previous invocation of
            L{VertexSeq.select()}. In this case, the indices do not refer
            directly to the vertices of the graph but to the elements of
            the filtered vertex sequence.

          - If the first positional argument is an integer, all remaining
            arguments are expected to be integers. They are considered as
            indices of the current vertex set again.

        Keyword arguments can be used to filter the vertices based on their
        attributes. The name of the keyword specifies the name of the attribute
        and the filtering operator, they should be concatenated by an
        underscore (C{_}) character. Attribute names can also contain
        underscores, but operator names don't, so the operator is always the
        largest trailing substring of the keyword name that does not contain
        an underscore. Possible operators are:

          - C{eq}: equal to

          - C{ne}: not equal to

          - C{lt}: less than

          - C{gt}: greater than

          - C{le}: less than or equal to

          - C{ge}: greater than or equal to

          - C{in}: checks if the value of an attribute is in a given list

          - C{notin}: checks if the value of an attribute is not in a given
            list

        For instance, if you want to filter vertices with a numeric C{age}
        property larger than 200, you have to write:

          >>> g.vs.select(age_gt=200)                   #doctest: +SKIP

        Similarly, to filter vertices whose C{type} is in a list of predefined
        types:

          >>> list_of_types = ["HR", "Finance", "Management"]
          >>> g.vs.select(type_in=list_of_types)        #doctest: +SKIP

        If the operator is omitted, it defaults to C{eq}. For instance, the
        following selector selects vertices whose C{cluster} property equals
        to 2:

          >>> g.vs.select(cluster=2)                    #doctest: +SKIP

        In the case of an unknown operator, it is assumed that the
        recognized operator is part of the attribute name and the actual
        operator is C{eq}.

        Attribute names inferred from keyword arguments are treated specially
        if they start with an underscore (C{_}). These are not real attributes
        but refer to specific properties of the vertices, e.g., its degree.
        The rule is as follows: if an attribute name starts with an underscore,
        the rest of the name is interpreted as a method of the L{Graph} object.
        This method is called with the vertex sequence as its first argument
        (all others left at default values) and vertices are filtered
        according to the value returned by the method. For instance, if you
        want to exclude isolated vertices:

          >>> g = Graph.Famous("zachary")
          >>> non_isolated = g.vs.select(_degree_gt=0)

        For properties that take a long time to be computed (e.g., betweenness
        centrality for large graphs), it is advised to calculate the values
        in advance and store it in a graph attribute. The same applies when
        you are selecting based on the same property more than once in the
        same C{select()} call to avoid calculating it twice unnecessarily.
        For instance, the following would calculate betweenness centralities
        twice:

          >>> edges = g.vs.select(_betweenness_gt=10, _betweenness_lt=30)

        It is advised to use this instead:

          >>> g.vs["bs"] = g.betweenness()
          >>> edges = g.vs.select(bs_gt=10, bs_lt=30)

        @return: the new, filtered vertex sequence"""
        vs = _VertexSeq.select(self, *args)

        operators = {
            "lt": operator.lt,
            "gt": operator.gt,
            "le": operator.le,
            "ge": operator.ge,
            "eq": operator.eq,
            "ne": operator.ne,
            "in": lambda a, b: a in b,
            "notin": lambda a, b: a not in b,
        }
        for keyword, value in kwds.items():
            if "_" not in keyword or keyword.rindex("_") == 0:
                keyword += "_eq"
            attr, _, op = keyword.rpartition("_")
            try:
                func = operators[op]
            except KeyError:
                # No such operator, assume that it's part of the attribute name
                attr, op, func = keyword, "eq", operators["eq"]

            if attr[0] == "_":
                # Method call, not an attribute
                values = getattr(vs.graph, attr[1:])(vs)
            else:
                values = vs[attr]
            filtered_idxs = [i for i, v in enumerate(values) if func(v, value)]
            vs = vs.select(filtered_idxs)

        return vs

    def __call__(self, *args, **kwds):
        """Shorthand notation to select()

        This method simply passes all its arguments to L{VertexSeq.select()}.
        """
        return self.select(*args, **kwds)


class EdgeSeq(_EdgeSeq):
    """Class representing a sequence of edges in the graph.

    This class is most easily accessed by the C{es} field of the
    L{Graph} object, which returns an ordered sequence of all edges in
    the graph. The edge sequence can be refined by invoking the
    L{EdgeSeq.select()} method. L{EdgeSeq.select()} can also be
    accessed by simply calling the L{EdgeSeq} object.

    An alternative way to create an edge sequence referring to a given
    graph is to use the constructor directly:

      >>> g = Graph.Full(3)
      >>> es = EdgeSeq(g)
      >>> restricted_es = EdgeSeq(g, [0, 1])

    The individual edges can be accessed by indexing the edge sequence
    object. It can be used as an iterable as well, or even in a list
    comprehension:

      >>> g=Graph.Full(3)
      >>> for e in g.es:
      ...   print(e.tuple)
      ...
      (0, 1)
      (0, 2)
      (1, 2)
      >>> [max(e.tuple) for e in g.es]
      [1, 2, 2]

    The edge sequence can also be used as a dictionary where the keys are the
    attribute names. The values corresponding to the keys are the values
    of the given attribute of every edge in the graph:

      >>> g=Graph.Full(3)
      >>> for idx, e in enumerate(g.es):
      ...   e["weight"] = idx*(idx+1)
      ...
      >>> g.es["weight"]
      [0, 2, 6]
      >>> g.es["weight"] = range(3)
      >>> g.es["weight"]
      [0, 1, 2]

    If you specify a sequence that is shorter than the number of edges in
    the EdgeSeq, the sequence is reused:

      >>> g = Graph.Tree(7, 2)
      >>> g.es["color"] = ["red", "green"]
      >>> g.es["color"]
      ['red', 'green', 'red', 'green', 'red', 'green']

    You can even pass a single string or integer, it will be considered as a
    sequence of length 1:

      >>> g.es["color"] = "red"
      >>> g.es["color"]
      ['red', 'red', 'red', 'red', 'red', 'red']

    Some methods of the edge sequences are simply proxy methods to the
    corresponding methods in the L{Graph} object. One such example is
    C{EdgeSeq.is_multiple()}:

      >>> g=Graph(3, [(0,1), (1,0), (1,2)])
      >>> g.es.is_multiple()
      [False, True, False]
      >>> g.es.is_multiple() == g.is_multiple()
      True
    """

    def attributes(self):
        """Returns the list of all the edge attributes in the graph
        associated to this edge sequence."""
        return self.graph.edge_attributes()

    def find(self, *args, **kwds):
        """Returns the first edge of the edge sequence that matches some
        criteria.

        The selection criteria are equal to the ones allowed by L{VertexSeq.select}.
        See L{VertexSeq.select} for more details.

        For instance, to find the first edge with weight larger than 5 in graph C{g}:

            >>> g.es.find(weight_gt=5)           #doctest:+SKIP
        """
        if args:
            # Selecting first based on positional arguments, then checking
            # the criteria specified by the keyword arguments
            edge = _EdgeSeq.find(self, *args)
            if not kwds:
                return edge
            es = self.graph.es.select(edge.index)
        else:
            es = self

        # Selecting based on positional arguments
        es = es.select(**kwds)
        if es:
            return es[0]
        raise ValueError("no such edge")

    def select(self, *args, **kwds):
        """Selects a subset of the edge sequence based on some criteria

        The selection criteria can be specified by the positional and the
        keyword arguments. Positional arguments are always processed before
        keyword arguments.

          - If the first positional argument is C{None}, an empty sequence is
            returned.

          - If the first positional argument is a callable object, the object
            will be called for every edge in the sequence. If it returns
            C{True}, the edge will be included, otherwise it will
            be excluded.

          - If the first positional argument is an iterable, it must return
            integers and they will be considered as indices of the current
            edge set (NOT the whole edge set of the graph -- the
            difference matters when one filters an edge set that has
            already been filtered by a previous invocation of
            L{EdgeSeq.select()}. In this case, the indices do not refer
            directly to the edges of the graph but to the elements of
            the filtered edge sequence.

          - If the first positional argument is an integer, all remaining
            arguments are expected to be integers. They are considered as
            indices of the current edge set again.

        Keyword arguments can be used to filter the edges based on their
        attributes and properties. The name of the keyword specifies the name
        of the attribute and the filtering operator, they should be
        concatenated by an underscore (C{_}) character. Attribute names can
        also contain underscores, but operator names don't, so the operator is
        always the largest trailing substring of the keyword name that does not
        contain an underscore. Possible operators are:

          - C{eq}: equal to

          - C{ne}: not equal to

          - C{lt}: less than

          - C{gt}: greater than

          - C{le}: less than or equal to

          - C{ge}: greater than or equal to

          - C{in}: checks if the value of an attribute is in a given list

          - C{notin}: checks if the value of an attribute is not in a given
            list

        For instance, if you want to filter edges with a numeric C{weight}
        property larger than 50, you have to write:

          >>> g.es.select(weight_gt=50)            #doctest: +SKIP

        Similarly, to filter edges whose C{type} is in a list of predefined
        types:

          >>> list_of_types = ["inhibitory", "excitatory"]
          >>> g.es.select(type_in=list_of_types)   #doctest: +SKIP

        If the operator is omitted, it defaults to C{eq}. For instance, the
        following selector selects edges whose C{type} property is
        C{intracluster}:

          >>> g.es.select(type="intracluster")     #doctest: +SKIP

        In the case of an unknown operator, it is assumed that the
        recognized operator is part of the attribute name and the actual
        operator is C{eq}.

        Keyword arguments are treated specially if they start with an
        underscore (C{_}). These are not real attributes but refer to specific
        properties of the edges, e.g., their centrality.  The rules are as
        follows:

          1. C{_source} or {_from} means the source vertex of an edge. For
             undirected graphs, only the C{eq} operator is supported and it
             is treated as {_incident} (since undirected graphs have no notion
             of edge directionality).

          2. C{_target} or {_to} means the target vertex of an edge. For
             undirected graphs, only the C{eq} operator is supported and it
             is treated as {_incident} (since undirected graphs have no notion
             of edge directionality).

          3. C{_within} ignores the operator and checks whether both endpoints
             of the edge lie within a specified set.

          4. C{_between} ignores the operator and checks whether I{one}
             endpoint of the edge lies within a specified set and the I{other}
             endpoint lies within another specified set. The two sets must be
             given as a tuple.

          5. C{_incident} ignores the operator and checks whether the edge is
             incident on a specific vertex or a set of vertices.

          6. Otherwise, the rest of the name is interpreted as a method of the
             L{Graph} object. This method is called with the edge sequence as
             its first argument (all others left at default values) and edges
             are filtered according to the value returned by the method.

        For instance, if you want to exclude edges with a betweenness
        centrality less than 2:

          >>> g = Graph.Famous("zachary")
          >>> excl = g.es.select(_edge_betweenness_ge = 2)

        To select edges originating from vertices 2 and 4:

          >>> edges = g.es.select(_source_in = [2, 4])

        To select edges lying entirely within the subgraph spanned by vertices
        2, 3, 4 and 7:

          >>> edges = g.es.select(_within = [2, 3, 4, 7])

        To select edges with one endpoint in the vertex set containing vertices
        2, 3, 4 and 7 and the other endpoint in the vertex set containing
        vertices 8 and 9:

          >>> edges = g.es.select(_between = ([2, 3, 4, 7], [8, 9]))

        For properties that take a long time to be computed (e.g., betweenness
        centrality for large graphs), it is advised to calculate the values
        in advance and store it in a graph attribute. The same applies when
        you are selecting based on the same property more than once in the
        same C{select()} call to avoid calculating it twice unnecessarily.
        For instance, the following would calculate betweenness centralities
        twice:

          >>> edges = g.es.select(_edge_betweenness_gt=10,       # doctest:+SKIP
          ...                     _edge_betweenness_lt=30)

        It is advised to use this instead:

          >>> g.es["bs"] = g.edge_betweenness()
          >>> edges = g.es.select(bs_gt=10, bs_lt=30)

        @return: the new, filtered edge sequence
        """
        es = _EdgeSeq.select(self, *args)
        is_directed = self.graph.is_directed()

        def _ensure_set(value):
            if isinstance(value, VertexSeq):
                value = {v.index for v in value}
            elif not isinstance(value, (set, frozenset)):
                value = set(value)
            return value

        operators = {
            "lt": operator.lt,
            "gt": operator.gt,
            "le": operator.le,
            "ge": operator.ge,
            "eq": operator.eq,
            "ne": operator.ne,
            "in": lambda a, b: a in b,
            "notin": lambda a, b: a not in b,
        }

        # TODO(ntamas): some keyword arguments should be prioritized over
        # others; for instance, we have optimized code paths for _source and
        # _target in directed and undirected graphs if es.is_all() is True;
        # these should be executed first. This matters only if there are
        # multiple keyword arguments and es.is_all() is True.

        for keyword, value in kwds.items():
            if "_" not in keyword or keyword.rindex("_") == 0:
                keyword += "_eq"
            pos = keyword.rindex("_")
            attr, op = keyword[0:pos], keyword[pos + 1 :]
            try:
                func = operators[op]
            except KeyError:
                # No such operator, assume that it's part of the attribute name
                attr, op, func = keyword, "eq", operators["eq"]

            if attr[0] == "_":
                if attr in ("_source", "_from", "_target", "_to") and not is_directed:
                    if op not in ("eq", "in"):
                        raise RuntimeError("unsupported for undirected graphs")

                    # translate to _incident to avoid confusion
                    attr = "_incident"
                    if func == operators["eq"]:
                        if hasattr(value, "__iter__") and not isinstance(value, str):
                            value = set(value)
                        else:
                            value = {value}

                if attr in ("_source", "_from"):
                    if es.is_all() and op == "eq":
                        # shortcut here: use .incident() as it is much faster
                        filtered_idxs = sorted(es.graph.incident(value, mode="out"))
                        func = None
                        # TODO(ntamas): there are more possibilities; we could
                        # optimize "ne", "in" and "notin" in similar ways
                    else:
                        values = [e.source for e in es]
                        if op == "in" or op == "notin":
                            value = _ensure_set(value)

                elif attr in ("_target", "_to"):
                    if es.is_all() and op == "eq":
                        # shortcut here: use .incident() as it is much faster
                        filtered_idxs = sorted(es.graph.incident(value, mode="in"))
                        func = None
                        # TODO(ntamas): there are more possibilities; we could
                        # optimize "ne", "in" and "notin" in similar ways
                    else:
                        values = [e.target for e in es]
                        if op == "in" or op == "notin":
                            value = _ensure_set(value)

                elif attr == "_incident":
                    func = None  # ignoring function, filtering here
                    value = _ensure_set(value)

                    # Fetch all the edges that are incident on at least one of
                    # the vertices specified
                    candidates = set()
                    for v in value:
                        candidates.update(es.graph.incident(v, mode="all"))

                    if not es.is_all():
                        # Find those that are in the current edge sequence
                        filtered_idxs = [
                            i for i, e in enumerate(es) if e.index in candidates
                        ]
                    else:
                        # We are done, the filtered indexes are in the candidates set
                        filtered_idxs = sorted(candidates)

                elif attr == "_within":
                    func = None  # ignoring function, filtering here
                    value = _ensure_set(value)

                    # Fetch all the edges that are incident on at least one of
                    # the vertices specified
                    candidates = set()
                    for v in value:
                        candidates.update(es.graph.incident(v))

                    if not es.is_all():
                        # Find those where both endpoints are OK
                        filtered_idxs = [
                            i
                            for i, e in enumerate(es)
                            if e.index in candidates
                            and e.source in value
                            and e.target in value
                        ]
                    else:
                        # Optimized version when the edge sequence contains all
                        # the edges exactly once in increasing order of edge IDs
                        filtered_idxs = [
                            i
                            for i in candidates
                            if es[i].source in value and es[i].target in value
                        ]

                elif attr == "_between":
                    if len(value) != 2:
                        raise ValueError(
                            "_between selector requires two vertex ID lists"
                        )
                    func = None  # ignoring function, filtering here
                    set1 = _ensure_set(value[0])
                    set2 = _ensure_set(value[1])

                    # Fetch all the edges that are incident on at least one of
                    # the vertices specified
                    candidates = set()
                    for v in set1:
                        candidates.update(es.graph.incident(v))
                    for v in set2:
                        candidates.update(es.graph.incident(v))

                    if not es.is_all():
                        # Find those where both endpoints are OK
                        filtered_idxs = [
                            i
                            for i, e in enumerate(es)
                            if (e.source in set1 and e.target in set2)
                            or (e.target in set1 and e.source in set2)
                        ]
                    else:
                        # Optimized version when the edge sequence contains all
                        # the edges exactly once in increasing order of edge IDs
                        filtered_idxs = [
                            i
                            for i in candidates
                            if (es[i].source in set1 and es[i].target in set2)
                            or (es[i].target in set1 and es[i].source in set2)
                        ]

                else:
                    # Method call, not an attribute
                    values = getattr(es.graph, attr[1:])(es)
            else:
                values = es[attr]

            # If we have a function to apply on the values, do that; otherwise
            # we assume that filtered_idxs has already been calculated.
            if func is not None:
                filtered_idxs = [i for i, v in enumerate(values) if func(v, value)]

            es = es.select(filtered_idxs)

        return es

    def __call__(self, *args, **kwds):
        """Shorthand notation to select()

        This method simply passes all its arguments to L{EdgeSeq.select()}.
        """
        return self.select(*args, **kwds)


def _graphmethod(func=None, name=None):
    """Auxiliary decorator

    This decorator allows some methods of L{VertexSeq} and L{EdgeSeq} to
    call their respective counterparts in L{Graph} to avoid code duplication.

    @param func: the function being decorated. This function will be
      called on the results of the original L{Graph} method.
      If C{None}, defaults to the identity function.
    @param name: the name of the corresponding method in L{Graph}. If
      C{None}, it defaults to the name of the decorated function.
    @return: the decorated function
    """
    # Delay import to avoid cycles
    from igraph import Graph

    if name is None:
        name = func.__name__
    method = getattr(Graph, name)

    if callable(func):

        def decorated(*args, **kwds):
            self = args[0].graph
            return func(args[0], method(self, *args, **kwds))

    else:

        def decorated(*args, **kwds):
            self = args[0].graph
            return method(self, *args, **kwds)

    decorated.__name__ = name
    decorated.__doc__ = """Proxy method to L{Graph.%(name)s()}

This method calls the C{%(name)s()} method of the L{Graph} class
restricted to this sequence, and returns the result.

@see: Graph.%(name)s() for details.
""" % {"name": name}

    return decorated


def _add_proxy_methods():
    # Proxy methods for VertexSeq and EdgeSeq that forward their arguments to
    # the corresponding Graph method are constructed here. Proxy methods for
    # Vertex and Edge are added in the C source code. Make sure that you update
    # the C source whenever you add a proxy method here if that makes sense for
    # an individual vertex or edge
    decorated_methods = {}
    decorated_methods[VertexSeq] = [
        "degree",
        "betweenness",
        "bibcoupling",
        "closeness",
        "cocitation",
        "constraint",
        "distances",
        "diversity",
        "eccentricity",
        "get_shortest_paths",
        "maxdegree",
        "pagerank",
        "personalized_pagerank",
        "shortest_paths",
        "similarity_dice",
        "similarity_jaccard",
        "subgraph",
        "indegree",
        "outdegree",
        "isoclass",
        "delete_vertices",
        "is_separator",
        "is_minimal_separator",
    ]
    decorated_methods[EdgeSeq] = [
        "count_multiple",
        "delete_edges",
        "is_loop",
        "is_multiple",
        "is_mutual",
        "subgraph_edges",
    ]

    rename_methods = {}
    rename_methods[VertexSeq] = {"delete_vertices": "delete"}
    rename_methods[EdgeSeq] = {"delete_edges": "delete", "subgraph_edges": "subgraph"}

    for cls, methods in decorated_methods.items():
        for method in methods:
            new_method_name = rename_methods[cls].get(method, method)
            setattr(cls, new_method_name, _graphmethod(None, method))

    EdgeSeq.edge_betweenness = _graphmethod(
        lambda self, result: [result[i] for i in self.indices], "edge_betweenness"
    )



# vim:ts=4 sw=4 sts=4:
import unittest
from igraph import Graph


class GraphAdjacencyMatrixLikeIndexingTests(unittest.TestCase):
    def testSingleEdgeRetrieval(self):
        g = Graph.Famous("krackhardt_kite")
        for v1, v2 in g.get_edgelist():
            self.assertEqual(g[v1, v2], 1)
            self.assertEqual(g[v2, v1], 1)
        for v1 in range(g.vcount()):
            for v2 in set(range(g.vcount())) - set(g.neighbors(v1)):
                self.assertEqual(g[v1, v2], 0)
                self.assertEqual(g[v2, v1], 0)

        g.add_edge(1, 1)
        self.assertEqual(g[1, 1], 1)

    def testSingleEdgeRetrievalWeights(self):
        g = Graph.Famous("krackhardt_kite")
        g.es["weight"] = list(range(g.ecount()))
        for idx, (v1, v2) in enumerate(g.get_edgelist()):
            self.assertEqual(g[v1, v2], idx)
            self.assertEqual(g[v2, v1], idx)
        for v1 in range(g.vcount()):
            for v2 in set(range(g.vcount())) - set(g.neighbors(v1)):
                self.assertEqual(g[v1, v2], 0)
                self.assertEqual(g[v2, v1], 0)

    def testSingleEdgeRetrievalAttrName(self):
        g = Graph.Famous("krackhardt_kite")
        g.es["value"] = list(range(20, g.ecount() + 20))
        for idx, (v1, v2) in enumerate(g.get_edgelist()):
            self.assertEqual(g[v1, v2, "value"], idx + 20)
            self.assertEqual(g[v2, v1, "value"], idx + 20)
        for v1 in range(g.vcount()):
            for v2 in set(range(g.vcount())) - set(g.neighbors(v1)):
                self.assertEqual(g[v1, v2, "value"], 0)
                self.assertEqual(g[v2, v1, "value"], 0)


def suite():
    adjacency_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        GraphAdjacencyMatrixLikeIndexingTests
    )
    return unittest.TestSuite([adjacency_suite])


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



"""This module provides implementation for a Cairo-specific histogram drawer"""

from igraph.drawing.cairo.base import AbstractCairoDrawer

__all__ = ("CairoHistogramDrawer",)


class CairoHistogramDrawer(AbstractCairoDrawer):
    """Default Cairo drawer object for histograms"""

    def __init__(self, context):
        """Constructs the vertex drawer and associates it to the given
        palette.

        @param context: the context on which we will draw
        """
        super().__init__(context, bbox=None)

    def draw(self, histogram, **kwds):
        """TODO"""
        from igraph.drawing.cairo.coord import DescartesCoordinateSystem

        context = self.context

        bbox = self.bbox = kwds.pop("bbox", None)
        if bbox is None:
            raise ValueError("bbox is required for Cairo plots")

        xmin = kwds.get("min", self._min)
        ymin = 0
        xmax = kwds.get("max", self._max)
        ymax = kwds.get("max_value", max(self._bins))
        width = self._bin_width

        coord_system = DescartesCoordinateSystem(
            context,
            bbox,
            (xmin, ymin, xmax, ymax),
        )

        # Draw the boxes
        context.set_line_width(1)
        context.set_source_rgb(1.0, 0.0, 0.0)
        x = self._min
        for value in self._bins:
            top_left_x, top_left_y = coord_system.local_to_context(x, value)
            x += width
            bottom_right_x, bottom_right_y = coord_system.local_to_context(x, 0)
            context.rectangle(
                top_left_x,
                top_left_y,
                bottom_right_x - top_left_x,
                bottom_right_y - top_left_y,
            )
            context.fill()

        # Draw the axes
        coord_system.draw()



#!usr/bin/env python

import os
import platform
import sys

###########################################################################

# Check Python's version info and exit early if it is too old
if sys.version_info < (3, 9):
    print("This module requires Python >= 3.9")
    sys.exit(0)

###########################################################################

from setuptools import find_packages, setup, Command, Extension

try:
    from setuptools.command.bdist_wheel import bdist_wheel
except ImportError:
    bdist_wheel = None

import glob
import shlex
import shutil
import subprocess
import sysconfig

from contextlib import contextmanager
from pathlib import Path
from select import select
from shutil import which
from time import sleep
from typing import List, Iterable, Iterator, Optional, Tuple, TypeVar, Union

###########################################################################

LIBIGRAPH_FALLBACK_INCLUDE_DIRS = ["/usr/include/igraph", "/usr/local/include/igraph"]
LIBIGRAPH_FALLBACK_LIBRARIES = ["igraph"]
LIBIGRAPH_FALLBACK_LIBRARY_DIRS = []

# Check whether we are compiling for PyPy or wasm with emscripten. Headers will
# not be installed in these cases, or when the SKIP_HEADER_INSTALL envvar is
# set explicitly.
SKIP_HEADER_INSTALL = (
    platform.python_implementation() == "PyPy"
    or (sysconfig.get_config_var("HOST_GNU_TYPE") or "").endswith("emscripten")
    or "SKIP_HEADER_INSTALL" in os.environ
)

###########################################################################


T = TypeVar("T")


def is_envvar_on(name: str) -> bool:
    """Returns whether the given environment variable is set to a truthy value
    such as '1', 'on' or 'true'.
    """
    value = os.environ.get(name, "")
    return value and str(value).lower() in ("1", "on", "true")


def building_on_windows_msvc() -> bool:
    """Returns True when using the non-MinGW CPython interpreter on Windows"""
    return platform.system() == "Windows" and sysconfig.get_platform() != "mingw"


def building_with_emscripten() -> bool:
    """Returns True when building with Emscripten to WebAssembly"""
    return (sysconfig.get_config_var("HOST_GNU_TYPE") or "").endswith("emscripten")


def building_with_sanitizers() -> bool:
    """Returns True when the IGRAPH_USE_SANITIZERS envvar is set, indicating that
    we want to build the Python interface with AddressSanitizer and LeakSanitizer
    enabled. Currently works on Linux only and the primary use-case is to be able
    to test igraph with sanitizers in CI.
    """
    return platform.system() == "Linux" and is_envvar_on("IGRAPH_USE_SANITIZERS")


def exclude_from_list(items: Iterable[T], items_to_exclude: Iterable[T]) -> List[T]:
    """Excludes certain items from a list, keeping the original order of
    the remaining items."""
    itemset = set(items_to_exclude)
    return [item for item in items if item not in itemset]


def fail(message: str, code: int = 1) -> None:
    """Fails the build with the given error message and exit code."""
    print(message)
    sys.exit(code)


def find_static_library(library_name: str, library_path: List[str]) -> Optional[str]:
    """Given the raw name of a library in `library_name`, tries to find a
    static library with this name in the given `library_path`. `library_path`
    is automatically extended with common library directories on Linux and Mac
    OS X."""

    variants = ["lib{0}.a", "{0}.a", "{0}.lib", "lib{0}.lib"]
    if is_unix_like():
        extra_libdirs = [
            "/opt/homebrew/lib",  # for newer Homebrew installations on macOS
            "/usr/local/lib64",
            "/usr/local/lib",
            "/usr/lib/x86_64-linux-gnu",
            "/usr/lib64",
            "/usr/lib",
            "/lib64",
            "/lib",
        ]
    else:
        extra_libdirs = []

    for path in extra_libdirs:
        if path not in library_path and os.path.isdir(path):
            library_path.append(path)

    for path in library_path:
        for variant in variants:
            full_path = os.path.join(path, variant.format(library_name))
            if os.path.isfile(full_path):
                return full_path


def first(iterable: Iterable[T]) -> T:
    """Returns the first element from the given iterable."""
    for item in iterable:
        return item
    raise ValueError("iterable is empty")


def get_output(args, encoding: str = "utf-8") -> Tuple[str, int]:
    """Returns the output of a command returning a single line of output, and
    the exit code of the command.
    """
    PIPE = subprocess.PIPE
    try:
        p = subprocess.Popen(args, shell=False, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate()
        returncode = p.returncode
    except OSError:
        stdout, stderr = None, None
        returncode = 77
    if isinstance(stdout, bytes):
        stdout = str(stdout, encoding=encoding)
    if isinstance(stderr, bytes):
        stderr = str(stderr, encoding=encoding)
    return (stdout or ""), returncode


def get_output_single_line(args, encoding: str = "utf-8") -> Tuple[str, int]:
    """Returns the first line of the output of a command, stripped from any
    trailing newlines, and the exit code of the command.
    """
    stdout, returncode = get_output(args, encoding=encoding)
    line, _, _ = stdout.partition("\n")
    return line, returncode


def is_unix_like(platform: str = sys.platform) -> bool:
    """Returns whether the given platform is a Unix-like platform with the usual
    Unix filesystem. When the parameter is omitted, it defaults to ``sys.platform``
    """
    platform = platform or sys.platform
    platform = platform.lower()
    return (
        platform.startswith("linux")
        or platform.startswith("darwin")
        or platform.startswith("cygwin")
    )


def wait_for_keypress(seconds: float) -> None:
    """Wait for a keypress or until the given number of seconds have passed,
    whichever happens first.
    """
    while seconds > 0:
        if seconds > 1:
            plural = "s"
        else:
            plural = ""

        sys.stdout.write(
            "\rContinuing in %2d second%s; press Enter to continue "
            "immediately. " % (seconds, plural)
        )
        sys.stdout.flush()

        if platform.system() == "Windows":
            from msvcrt import kbhit  # type: ignore

            for _ in range(10):
                if kbhit():
                    seconds = 0
                    break
                sleep(0.1)
        else:
            rlist, _, _ = select([sys.stdin], [], [], 1)
            if rlist:
                sys.stdin.readline()
                seconds = 0
                break

        seconds -= 1

    sys.stdout.write("\r" + " " * 65 + "\r")


@contextmanager
def working_directory(dir: Union[str, Path]) -> Iterator[None]:
    cwd = os.getcwd()
    os.chdir(dir)
    try:
        yield
    finally:
        os.chdir(cwd)


###########################################################################


class IgraphCCoreCMakeBuilder:
    """Class responsible for downloading and building the C core of igraph
    if it is not installed yet, assuming that the C core uses CMake as the
    build tool. This is the case from igraph 0.9.
    """

    def compile_in(
        self, source_folder: Path, build_folder: Path, install_folder: Path
    ) -> Union[bool, List[str]]:
        """Compiles igraph from its source code in the given folder.

        Parameters:
            source_folder: absolute path to the folder that contains igraph's
                source files
            build_folder: absolute path to the folder where the build should be
                executed
            install_folder: absolute path to the folder where the built library
                should be installed

        Returns:
            False if the build failed or the list of libraries to link to when
            linking the Python interface to igraph
        """
        with working_directory(build_folder):
            return self._compile_in(source_folder, build_folder, install_folder)

    def _compile_in(
        self, source_folder: Path, build_folder: Path, install_folder: Path
    ) -> Union[bool, List[str]]:
        cmake = which("cmake")
        if not cmake:
            print(
                "igraph uses CMake as the build system. You need to install CMake "
                "before compiling igraph."
            )
            return False

        build_to_source_folder = os.path.relpath(source_folder, build_folder)

        print("Configuring build...")
        args = [cmake]
        cmake_build_mode = "Release"

        # Build to wasm requires invocation of the Emscripten SDK
        if building_with_emscripten():
            emcmake = which("emcmake")
            if not emcmake:
                print(
                    "You need to install emcmake from the Emscripten SDK before "
                    "compiling igraph."
                )
                return False
            args.insert(0, emcmake)
            args.append("-DIGRAPH_WARNINGS_AS_ERRORS:BOOL=OFF")
            args.append("-DIGRAPH_GRAPHML_SUPPORT:BOOL=OFF")

        # Build the Python interface with vendored libraries
        for deps in "ARPACK BLAS GLPK GMP LAPACK PLFIT".split():
            args.append("-DIGRAPH_USE_INTERNAL_" + deps + "=ON")

        # Use link-time optinization if available
        args.append("-DIGRAPH_ENABLE_LTO=AUTO")

        # -fPIC is needed on Linux so we can link to a static igraph lib from a
        # Python shared library
        args.append("-DCMAKE_POSITION_INDEPENDENT_CODE=ON")

        # No need to build tests
        args.append("-DBUILD_TESTING=OFF")

        # Do not treat compilation warnings as errors in case someone is trying
        # to "pip install" igraph in an environment for which we don't provide
        # wheels and the compiler complains about harmless things
        args.append("-DIGRAPH_WARNINGS_AS_ERRORS=OFF")

        # Set install directory during config step instead of install step in order
        # to avoid having the architecture name in the LIBPATH (e.g. lib/x86_64-linux-gnu)
        args.append("-DCMAKE_INSTALL_PREFIX=" + str(install_folder))

        # On macOS, compile the C core with the same macOS deployment target as
        # the one that was used to compile Python itself
        if sysconfig.get_config_var("MACOSX_DEPLOYMENT_TARGET"):
            args.append(
                "-DCMAKE_OSX_DEPLOYMENT_TARGET="
                + sysconfig.get_config_var("MACOSX_DEPLOYMENT_TARGET")
            )

        # Compile the C core with sanitizers if needed
        if building_with_sanitizers():
            args.append("-DUSE_SANITIZER=Address;Undefined")
            args.append("-DFLEX_KEEP_LINE_NUMBERS=ON")
            cmake_build_mode = "Debug"

        # Add any extra CMake args from environment variables
        if "IGRAPH_CMAKE_EXTRA_ARGS" in os.environ:
            args.extend(
                shlex.split(
                    os.environ["IGRAPH_CMAKE_EXTRA_ARGS"],
                    posix=not building_on_windows_msvc(),
                )
            )

        # Finally, add the source folder path
        args.append(str(build_to_source_folder))

        retcode = subprocess.call(args)
        if retcode:
            return False

        print("Running build...")
        # We are _not_ using a parallel build; this is intentional, see igraph/igraph#1755
        retcode = subprocess.call([cmake, "--build", ".", "--config", cmake_build_mode])
        if retcode:
            return False

        print("Installing build...")
        retcode = subprocess.call(
            [
                cmake,
                "--install",
                ".",
                "--config",
                cmake_build_mode,
            ]
        )
        if retcode:
            return False

        for candidate in install_folder.rglob("igraph.pc"):
            return self._parse_pkgconfig_file(candidate)

        raise RuntimeError(
            "no igraph.pc was found in the installation folder of igraph"
        )

    def create_build_config_file(
        self, install_folder: Path, libraries: List[str]
    ) -> None:
        with (install_folder / "build.cfg").open("w") as fp:
            fp.write(repr(libraries))

    def _parse_pkgconfig_file(self, filename: Path) -> List[str]:
        building_on_windows = building_on_windows_msvc()

        if building_on_windows:
            libraries = ["igraph"]
        else:
            libraries = []
            with filename.open("r") as fp:
                for line in fp:
                    if line.startswith("Libs: ") or line.startswith("Libs.private: "):
                        words = line.strip().split()
                        libraries.extend(
                            word[2:] for word in words if word.startswith("-l")
                        )
                    # Remap known library names in Requires and Requires.private with
                    # prior knowledge -- we don't want to rebuild pkg-config in Python
                    if line.startswith("Requires: ") or line.startswith(
                        "Requires.private: "
                    ):
                        for word in line.strip().split():
                            if word.startswith("libxml-"):
                                libraries.append("xml2")
            if not libraries:
                # Educated guess
                libraries = ["igraph"]

        return libraries


###########################################################################


class BuildConfiguration:
    def __init__(self):
        self.include_dirs = []
        self.library_dirs = []
        self.runtime_library_dirs = []
        self.libraries = []
        self.extra_compile_args = []
        self.extra_link_args = []
        self.define_macros = []
        self.extra_objects = []
        self.static_extension = False
        self.use_pkgconfig = False
        self.use_sanitizers = False
        self.c_core_built = False
        self.allow_educated_guess = True
        self._has_pkgconfig = None
        self.excluded_include_dirs = []
        self.excluded_library_dirs = []
        self.wait = platform.system() != "Windows"

    @property
    def has_pkgconfig(self) -> bool:
        """Returns whether ``pkg-config`` is available on the current system
        and it knows about igraph or not."""
        if self._has_pkgconfig is None:
            if self.use_pkgconfig:
                _, exit_code = get_output_single_line(["pkg-config", "igraph"])
                self._has_pkgconfig = exit_code == 0
            else:
                self._has_pkgconfig = False
        return self._has_pkgconfig

    @property
    def build_c_core(self) -> Command:
        """Returns a class representing a custom setup.py command that builds
        the C core of igraph.

        This is used in CI environments where we want to build the C core of
        igraph once and then build the Python interface for various Python
        versions without having to recompile the C core all the time.

        If is also used as a custom building block of `build_ext`.
        """

        buildcfg = self

        class build_c_core(Command):
            description = "Compile the C core of igraph only"
            user_options = []

            def initialize_options(self):
                pass

            def finalize_options(self):
                pass

            def run(self):
                buildcfg.c_core_built = buildcfg.compile_igraph_from_vendor_source()

        return build_c_core

    @property
    def build_ext(self) -> Command:
        """Returns a class that can be used as a replacement for the
        ``build_ext`` command in ``setuptools`` and that will compile the C core
        of igraph before compiling the Python extension.
        """
        from setuptools.command.build_ext import build_ext

        buildcfg = self

        class custom_build_ext(build_ext):
            def run(self):
                # Bail out if we don't have the Python include files
                include_dir = sysconfig.get_path("include")
                if not os.path.isfile(os.path.join(include_dir, "Python.h")):
                    fail("You will need the Python headers to compile this extension.")

                # Check whether the user asked us to discover a pre-built igraph
                # with pkg-config
                detected = False
                if buildcfg.use_pkgconfig:
                    detected = buildcfg.detect_from_pkgconfig()
                    if not detected:
                        fail(
                            "Cannot find the C core of igraph on this system using pkg-config."
                        )
                else:
                    # Build the C core from the vendored igraph source
                    self.run_command("build_c_core")
                    if not buildcfg.c_core_built:
                        # Fall back to an educated guess if everything else failed
                        if not detected:
                            if buildcfg.allow_educated_guess:
                                buildcfg.use_educated_guess()
                            else:
                                fail("Cannot build the C core of igraph.")

                # Add any extra library paths if needed; this is needed for the
                # Appveyor CI build
                if "IGRAPH_EXTRA_LIBRARY_PATH" in os.environ:
                    buildcfg.library_dirs = (
                        list(os.environ["IGRAPH_EXTRA_LIBRARY_PATH"].split(os.pathsep))
                        + buildcfg.library_dirs
                    )

                # Add extra libraries that may have been specified
                if "IGRAPH_EXTRA_LIBRARIES" in os.environ:
                    extra_libraries = os.environ["IGRAPH_EXTRA_LIBRARIES"].split(",")
                    buildcfg.libraries.extend(extra_libraries)

                # Override build configuration based on environment variables
                if "IGRAPH_STATIC_EXTENSION" in os.environ:
                    buildcfg.static_extension = is_envvar_on("IGRAPH_STATIC_EXTENSION")
                buildcfg.use_sanitizers = building_with_sanitizers()

                # Replaces library names with full paths to static libraries
                # where possible. libm.a is excluded because it caused problems
                # on Sabayon Linux where libm.a is probably not compiled with
                # -fPIC
                if buildcfg.static_extension:
                    if buildcfg.static_extension == "only_igraph":
                        buildcfg.replace_static_libraries(only=["igraph"])
                    else:
                        buildcfg.replace_static_libraries(exclusions=["m"])

                # Add sanitizer flags
                if buildcfg.use_sanitizers:
                    buildcfg.extra_link_args += [
                        "-fsanitize=address",
                        "-fsanitize=undefined",
                    ]
                    buildcfg.extra_compile_args += [
                        "-g",
                        "-Og",
                        "-fno-omit-frame-pointer",
                        "-fdiagnostics-color",
                    ]

                # Add extra libraries that may have been specified
                if "IGRAPH_EXTRA_DYNAMIC_LIBRARIES" in os.environ:
                    extra_libraries = os.environ[
                        "IGRAPH_EXTRA_DYNAMIC_LIBRARIES"
                    ].split(",")
                    buildcfg.libraries.extend(extra_libraries)

                # Remove C++ standard library as we will use the C++ linker
                for lib in ("c++", "stdc++"):
                    if lib in buildcfg.libraries:
                        buildcfg.libraries.remove(lib)

                # Prints basic build information
                buildcfg.print_build_info()

                # Find the igraph extension and configure it with the settings
                # of this build configuration
                ext = first(
                    extension
                    for extension in self.extensions
                    if extension.name == "igraph._igraph"
                )
                buildcfg.configure(ext)

                # Run the original build_ext command
                build_ext.run(self)

        return custom_build_ext

    @property
    def sdist(self):
        """Returns a class that can be used as a replacement for the
        ``sdist`` command in ``setuptools`` and that will clean up
        ``vendor/source/igraph`` before running the original ``sdist``
        command.
        """
        from setuptools.command.sdist import sdist

        def is_git_repo(folder) -> bool:
            return (Path(folder) / ".git").exists()

        def cleanup_git_repo(folder) -> None:
            with working_directory(folder):
                if os.path.exists(".git"):
                    retcode = subprocess.call("git clean -dfx", shell=True)
                    if retcode:
                        raise RuntimeError(f"Failed to clean {folder} with git")

        class custom_sdist(sdist):
            def run(self):
                igraph_source_repo = Path("vendor", "source", "igraph")
                igraph_build_dir = Path("vendor", "build", "igraph")
                version_file = igraph_source_repo / "IGRAPH_VERSION"
                version = None

                # Check whether the source repo contains an IGRAPH_VERSION file,
                # and extract the version number from that
                if version_file.exists():
                    version = version_file.read_text().strip().split("\n")[0]

                # If no IGRAPH_VERSION file exists, but we have a git repo, try
                # git describe
                if not version and is_git_repo(igraph_source_repo):
                    with working_directory(igraph_source_repo):
                        version = (
                            subprocess.check_output("git describe", shell=True)
                            .decode("utf-8")
                            .strip()
                        )

                # If we still don't have a version number, try to parse it from
                # include/igraph_version.h
                if not version:
                    version_header = igraph_build_dir / "include" / "igraph_version.h"
                    if not version_header.exists():
                        raise RuntimeError(
                            "You need to build the C core of igraph first before generating a source tarball of the Python interface of igraph"
                        )

                    with version_header.open("r") as fp:
                        lines = [
                            line.strip()
                            for line in fp
                            if line.startswith("#define IGRAPH_VERSION ")
                        ]
                        if len(lines) == 1:
                            version = lines[0].split('"')[1]

                if not isinstance(version, str) or len(version) < 5:
                    raise RuntimeError(
                        f"Cannot determine the version number of the C core in {igraph_source_repo}"
                    )

                if not is_git_repo(igraph_source_repo):
                    # The Python interface was extracted from an official
                    # tarball so there is no need to tweak anything
                    return sdist.run(self)
                else:
                    # Clean up vendor/source/igraph with git
                    cleanup_git_repo(igraph_source_repo)

                    # Copy the generated parser sources from the build folder
                    parser_dir = igraph_build_dir / "src" / "io" / "parsers"
                    if parser_dir.is_dir():
                        shutil.copytree(
                            parser_dir, igraph_source_repo / "src" / "io" / "parsers"
                        )
                    else:
                        raise RuntimeError(
                            "You need to build the C core of igraph first before "
                            "generating a source tarball of the Python interface"
                        )

                    # Add a version file to the tarball
                    version_file.write_text(version)

                    # Run the original sdist command
                    retval = sdist.run(self)

                    # Clean up vendor/source/igraph with git again
                    cleanup_git_repo(igraph_source_repo)

                    return retval

        return custom_sdist

    def compile_igraph_from_vendor_source(self) -> bool:
        """Compiles igraph from the vendored source code inside `vendor/source/igraph`.
        This folder typically comes from a git submodule.
        """
        vendor_folder = Path("vendor")
        source_folder = vendor_folder / "source" / "igraph"
        build_folder = vendor_folder / "build" / "igraph"
        install_folder = vendor_folder / "install" / "igraph"

        if install_folder.exists():
            # Vendored igraph already compiled and installed, just use it
            self.use_vendored_igraph()
            return True

        if (source_folder / "CMakeLists.txt").exists():
            igraph_builder = IgraphCCoreCMakeBuilder()
        else:
            print("Cannot find vendored igraph source in {0}".format(source_folder))
            print("")
            return False

        print("We are going to build the C core of igraph.")
        print("  Source folder: {0}".format(source_folder))
        print("  Build folder: {0}".format(build_folder))
        print("  Install folder: {0}".format(install_folder))
        print("")

        source_folder = source_folder.resolve()
        build_folder = build_folder.resolve()
        install_folder = install_folder.resolve()

        Path(build_folder).mkdir(parents=True, exist_ok=True)

        libraries = igraph_builder.compile_in(
            source_folder=source_folder,
            build_folder=build_folder,
            install_folder=install_folder,
        )

        if libraries is False:
            fail("Build failed for the C core of igraph.")

        assert not isinstance(libraries, bool)

        igraph_builder.create_build_config_file(install_folder, libraries)

        self.use_vendored_igraph()
        return True

    def configure(self, ext) -> None:
        """Configures the given Extension object using this build configuration."""
        ext.include_dirs = exclude_from_list(
            self.include_dirs, self.excluded_include_dirs
        )
        ext.library_dirs = exclude_from_list(
            self.library_dirs, self.excluded_library_dirs
        )
        ext.runtime_library_dirs = self.runtime_library_dirs
        ext.libraries = self.libraries
        ext.extra_compile_args = self.extra_compile_args
        ext.extra_link_args = self.extra_link_args
        ext.extra_objects = self.extra_objects
        ext.define_macros = self.define_macros

    def detect_from_pkgconfig(self) -> bool:
        """Detects the igraph include directory, library directory and the
        list of libraries to link to using ``pkg-config``."""
        if not buildcfg.has_pkgconfig:
            return False

        cmd = ["pkg-config", "igraph", "--cflags", "--libs"]
        if self.static_extension:
            cmd += ["--static"]
        line, exit_code = get_output_single_line(cmd)
        if exit_code > 0 or len(line) == 0:
            return False

        opts = line.strip().split()
        self.libraries = [opt[2:] for opt in opts if opt.startswith("-l")]
        self.library_dirs = [opt[2:] for opt in opts if opt.startswith("-L")]
        self.include_dirs = [opt[2:] for opt in opts if opt.startswith("-I")]
        return True

    def print_build_info(self) -> None:
        """Prints the include and library path being used for debugging purposes."""
        if self.static_extension == "only_igraph":
            build_type = "dynamic extension with vendored igraph source"
        elif self.static_extension:
            build_type = "static extension"
        else:
            build_type = "dynamic extension"
        print("Build type: %s" % build_type)
        print("Include path: %s" % " ".join(self.include_dirs))
        if self.excluded_include_dirs:
            print("  - excluding: %s" % " ".join(self.excluded_include_dirs))
        print("Library path: %s" % " ".join(self.library_dirs))
        if self.excluded_library_dirs:
            print("  - excluding: %s" % " ".join(self.excluded_library_dirs))
        print("Runtime library path: %s" % " ".join(self.runtime_library_dirs))
        print("Linked dynamic libraries: %s" % " ".join(self.libraries))
        print("Linked static libraries: %s" % " ".join(self.extra_objects))
        print("Extra compiler options: %s" % " ".join(self.extra_compile_args))
        print("Extra linker options: %s" % " ".join(self.extra_link_args))

    def process_args_from_command_line(self):
        """Preprocesses the command line options before they are passed to
        setup.py and sets up the build configuration.
        """
        # Yes, this is ugly, but we don't want to interfere with setup.py's own
        # option handling
        opts_to_remove = []
        for idx, option in enumerate(sys.argv):
            if not option.startswith("--"):
                continue
            if option == "--static":
                opts_to_remove.append(idx)
                self.static_extension = True
            elif option == "--no-pkg-config":
                opts_to_remove.append(idx)
                self.use_pkgconfig = False
            elif option == "--no-wait":
                opts_to_remove.append(idx)
                self.wait = False
            elif option == "--use-pkg-config":
                opts_to_remove.append(idx)
                self.use_pkgconfig = True

        for idx in reversed(opts_to_remove):
            sys.argv[idx : (idx + 1)] = []

    def process_environment_variables(self):
        """Processes environment variables that serve as replacements for the
        command line options. This is typically useful in CI environments where
        it is easier to set up a few environment variables permanently than to
        pass the same options to ``setup.py build`` and ``setup.py install``
        at the same time.
        """

        def process_envvar(name, attr, value):
            name = "IGRAPH_" + name.upper()
            if name in os.environ:
                value = str(os.environ[name]).lower()
                if value in ("on", "true", "yes"):
                    value = True
                elif value in ("off", "false", "no"):
                    value = False
                else:
                    try:
                        value = bool(int(value))
                    except Exception:
                        return

                setattr(self, attr, value)

        process_envvar("static", "static_extension", True)
        process_envvar("no_pkg_config", "use_pkgconfig", False)
        process_envvar("no_wait", "wait", False)
        process_envvar("use_pkg_config", "use_pkgconfig", True)
        process_envvar("use_sanitizers", "use_sanitizers", False)

    def replace_static_libraries(self, only=None, exclusions=None):
        """Replaces references to libraries with full paths to their static
        versions if the static version is to be found on the library path."""
        if exclusions is None:
            exclusions = []

        for library_name in set(self.libraries) - set(exclusions):
            if only is not None and library_name not in only:
                continue

            static_lib = find_static_library(library_name, self.library_dirs)
            if static_lib:
                print(f"Found {library_name} as static library in {static_lib}.")
                self.libraries.remove(library_name)
                self.extra_objects.append(static_lib)
            else:
                print(f"Warning: could not find static library of {library_name}.")

    def use_vendored_igraph(self) -> None:
        """Assumes that igraph is installed already in ``vendor/install/igraph`` and sets up
        the include and library paths and the library names accordingly."""
        building_on_windows = building_on_windows_msvc()

        vendor_dir = Path("vendor") / "install" / "igraph"

        buildcfg.include_dirs = [str(vendor_dir / "include" / "igraph")]
        buildcfg.library_dirs = []

        for candidate in ("lib", "lib64"):
            candidate = vendor_dir / candidate
            if candidate.exists():
                buildcfg.library_dirs.append(str(candidate))
                break
        else:
            raise RuntimeError(
                "cannot detect igraph library dir within " + str(vendor_dir)
            )

        if not buildcfg.static_extension:
            buildcfg.static_extension = "only_igraph"
            if building_on_windows:
                buildcfg.define_macros.append(("IGRAPH_STATIC", "1"))

        buildcfg_file = vendor_dir / "build.cfg"
        if buildcfg_file.exists():
            buildcfg.libraries = eval(buildcfg_file.open("r").read())

    def use_educated_guess(self) -> None:
        """Tries to guess the proper library names, include and library paths
        if everything else failed."""

        global LIBIGRAPH_FALLBACK_LIBRARIES
        global LIBIGRAPH_FALLBACK_INCLUDE_DIRS
        global LIBIGRAPH_FALLBACK_LIBRARY_DIRS

        print("WARNING: we were not able to detect where igraph is installed on")
        print("your machine (if it is installed at all). We will use the fallback")
        print("library and include paths hardcoded in setup.py and hope that the")
        print("C core of igraph is installed there.")
        print("")
        print("If the compilation fails and you are sure that igraph is installed")
        print("on your machine, adjust the following two variables in setup.py")
        print("accordingly and try again:")
        print("")
        print("- LIBIGRAPH_FALLBACK_INCLUDE_DIRS")
        print("- LIBIGRAPH_FALLBACK_LIBRARY_DIRS")
        print("")

        if self.wait:
            wait_for_keypress(seconds=10)

        self.libraries = LIBIGRAPH_FALLBACK_LIBRARIES[:]
        if self.static_extension:
            self.libraries.extend(["xml2", "z", "m", "stdc++"])
        self.include_dirs = LIBIGRAPH_FALLBACK_INCLUDE_DIRS[:]
        self.library_dirs = LIBIGRAPH_FALLBACK_LIBRARY_DIRS[:]


###########################################################################

if bdist_wheel is not None:

    class bdist_wheel_abi3(bdist_wheel):
        def get_tag(self):
            python, abi, plat = super().get_tag()
            if python.startswith("cp"):
                # on CPython, our wheels are abi3 and compatible back to 3.9
                return "cp39", "abi3", plat

            return python, abi, plat

else:
    bdist_wheel_abi3 = None

# We are going to build an abi3 wheel if we are at least on CPython 3.9.
should_build_abi3_wheel = (
    bdist_wheel_abi3
    and platform.python_implementation() == "CPython"
    and sys.version_info >= (3, 9)
)

###########################################################################

# Import version number from version.py so we only need to change it in
# one place when a new release is created
__version__: str = ""
exec(open("src/igraph/version.py").read())

# Process command line options
buildcfg = BuildConfiguration()
buildcfg.process_environment_variables()
buildcfg.process_args_from_command_line()

# Define the extension
sources = glob.glob(os.path.join("src", "_igraph", "*.c"))
sources.append(os.path.join("src", "_igraph", "force_cpp_linker.cpp"))
macros = []
if should_build_abi3_wheel:
    macros.append(("Py_LIMITED_API", "0x03090000"))
igraph_extension = Extension(
    "igraph._igraph",
    sources=sources,
    py_limited_api=should_build_abi3_wheel,
    define_macros=macros,
)

description = """Python interface to the igraph high performance graph
library, primarily aimed at complex network research and analysis.

Graph plotting functionality is provided by the Cairo library, so make
sure you install the Python bindings of Cairo if you want to generate
publication-quality graph plots. You can try either `pycairo
<http://cairographics.org/pycairo>`_ or `cairocffi <https://doc.courtbouillon.org/cairocffi/>`_,
``cairocffi`` is recommended because there were bug reports affecting igraph
graph plots in Jupyter notebooks when using ``pycairo`` (but not with
``cairocffi``).
"""

headers = ["src/_igraph/igraphmodule_api.h"] if not SKIP_HEADER_INSTALL else []

cmdclass = {
    "build_c_core": buildcfg.build_c_core,  # used by CI
    "build_ext": buildcfg.build_ext,
    "sdist": buildcfg.sdist,
}

if should_build_abi3_wheel:
    cmdclass["bdist_wheel"] = bdist_wheel_abi3

options = {
    "name": "igraph",
    "version": __version__,
    "url": "https://igraph.org/python",
    "description": "High performance graph data structures and algorithms",
    "long_description": description,
    "license": "GNU General Public License (GPL)",
    "author": "Tamas Nepusz",
    "author_email": "ntamas@gmail.com",
    "project_urls": {
        "Bug Tracker": "https://github.com/igraph/python-igraph/issues",
        "Changelog": "https://github.com/igraph/python-igraph/blob/main/CHANGELOG.md",
        "CI": "https://github.com/igraph/python-igraph/actions",
        "Documentation": "https://python.igraph.org",
        "Source Code": "https://github.com/igraph/python-igraph",
    },
    "ext_modules": [igraph_extension],
    "package_dir": {
        # make sure to use the next line and not the more logical and restrictive
        # "igraph": "src/igraph" because that one breaks 'setup.py develop'.
        # See: https://github.com/igraph/python-igraph/issues/464
        "": "src"
    },
    "packages": find_packages(where="src"),
    "install_requires": ["texttable>=1.6.2"],
    "entry_points": {"console_scripts": ["igraph=igraph.app.shell:main"]},
    "extras_require": {
        # Dependencies needed for plotting with Cairo
        "cairo": ["cairocffi>=1.2.0"],
        # Dependencies needed for plotting with Matplotlib
        "matplotlib": ["matplotlib>=3.6.0; platform_python_implementation != 'PyPy'"],
        # Dependencies needed for plotting with Plotly
        "plotly": ["plotly>=5.3.0"],
        # Compatibility alias to 'cairo' for python-igraph <= 0.9.6
        "plotting": ["cairocffi>=1.2.0"],
        # Dependencies needed for testing only
        "test": [
            "cairocffi>=1.2.0",
            "networkx>=2.5",
            "pytest>=7.0.1",
            "pytest-timeout>=2.1.0",
            "numpy>=1.19.0; platform_python_implementation != 'PyPy'",
            "pandas>=1.1.0; platform_python_implementation != 'PyPy'",
            "scipy>=1.5.0; platform_python_implementation != 'PyPy'",
            "matplotlib>=3.6.0; platform_python_implementation != 'PyPy'",
            "plotly>=5.3.0",
            "Pillow>=9; platform_python_implementation != 'PyPy'",
        ],
        # Dependencies needed for testing on Windows ARM64; only those that are either
        # pure Python or have Windows ARM64 wheels as we don't want to compile wheels
        # in CI
        "test-win-arm64": [
            "cairocffi>=1.2.0",
            "networkx>=2.5",
            "pytest>=7.0.1",
            "pytest-timeout>=2.1.0",
        ],
        # Dependencies needed for testing on musllinux; only those that are either
        # pure Python or have musllinux wheels as we don't want to compile wheels
        # in CI
        "test-musl": [
            "cairocffi>=1.2.0",
            "networkx>=2.5",
            "pytest>=7.0.1",
            "pytest-timeout>=2.1.0",
        ],
        # Dependencies needed for building the documentation
        "doc": [
            "Sphinx>=7.0.0",
            "sphinx-rtd-theme>=1.3.0",
            "sphinx-gallery>=0.14.0",
            "pydoctor>=23.4.0",
        ],
    },
    "python_requires": ">=3.9",
    "headers": headers,
    "platforms": "ALL",
    "keywords": [
        "graph",
        "network",
        "mathematics",
        "math",
        "graph theory",
        "discrete mathematics",
    ],
    "classifiers": [
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: C",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    "cmdclass": cmdclass,
}

setup(**options)



"""This module provides implementation for a Matplotlib-specific matrix drawer."""

from igraph.drawing.baseclasses import AbstractDrawer

__all__ = ("MatplotlibMatrixDrawer",)


class MatplotlibMatrixDrawer(AbstractDrawer):
    """Matplotlib drawer object for matrices."""

    def __init__(self, ax):
        """Constructs the drawer and associates it to the given Axes.

        @param ax: the Axes on which we will draw
        """
        self.context = ax

    def draw(self, matrix, **kwds):
        """Draws the given Matrix in a matplotlib Axes.

        @param matrix: the igraph.Matrix to plot.

        Keyword arguments are passed to Axes.imshow.
        """
        ax = self.context
        ax.imshow(matrix.data, interpolation="nearest", **kwds)



"""
.. _tutorials-online-user-actions:

===================
Online user actions
===================

This example reproduces a typical data science situation in an internet company. We start from a pandas DataFrame with online user actions, for instance for an online text editor: the user can create a page, edit it, or delete it. We want to construct and visualize a graph of the users highlighting collaborations on the same page/project.
"""

import igraph as ig
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

# %%
# Let's start by preparing some toy data representing online users. Each row
# indicates a certain action taken by a user (e.g. click on a button within a
# website). Actual user data usually come with time stamp, but that's not
# essential for this example.
action_dataframe = pd.DataFrame(
    [
        ["dsj3239asadsa3", "createPage", "greatProject"],
        ["2r09ej221sk2k5", "editPage", "greatProject"],
        ["dsj3239asadsa3", "editPage", "greatProject"],
        ["789dsadafj32jj", "editPage", "greatProject"],
        ["oi32ncwosap399", "editPage", "greatProject"],
        ["4r4320dkqpdokk", "createPage", "miniProject"],
        ["320eljl3lk3239", "editPage", "miniProject"],
        ["dsj3239asadsa3", "editPage", "miniProject"],
        ["3203ejew332323", "createPage", "private"],
        ["3203ejew332323", "editPage", "private"],
        ["40m11919332msa", "createPage", "private2"],
        ["40m11919332msa", "editPage", "private2"],
        ["dsj3239asadsa3", "createPage", "anotherGreatProject"],
        ["2r09ej221sk2k5", "editPage", "anotherGreatProject"],
    ],
    columns=["userid", "action", "project"],
)

# %%
# The goal of this example is to check when two users worked on the same page.
# We choose to use a weighted adjacency matrix for this, i.e. a table with rows
# and columns indexes by the users that has nonzero entries whenever folks
# collaborate. First, let's get the users and prepare an empty matrix:
users = action_dataframe["userid"].unique()
adjacency_matrix = pd.DataFrame(
    np.zeros((len(users), len(users)), np.int32),
    index=users,
    columns=users,
)

# %%
# Then, let's iterate over all projects one by one, and add all collaborations:
for _project, project_data in action_dataframe.groupby("project"):
    project_users = project_data["userid"].values
    for i1, user1 in enumerate(project_users):
        for user2 in project_users[:i1]:
            adjacency_matrix.at[user1, user2] += 1

# %%
# There are many ways to achieve the above matrix, so don't be surprised if you
# came up with another algorithm ;-) Now it's time to make the graph:
g = ig.Graph.Weighted_Adjacency(adjacency_matrix, mode="plus")

# %%
# We can take a look at the graph via plotting functions. We can first make a
# layout:
layout = g.layout("circle")

# %%
# Then we can prepare vertex sizes based on their closeness to other vertices
vertex_size = g.closeness()
vertex_size = [10 * v**2 if not np.isnan(v) else 10 for v in vertex_size]

# %%
# Finally, we can plot the graph:
fig, ax = plt.subplots()
ig.plot(
    g,
    target=ax,
    layout=layout,
    vertex_label=g.vs["name"],
    vertex_color="lightblue",
    vertex_size=vertex_size,
    edge_width=g.es["weight"],
)
plt.show()

# %%
# Loops indicate "self-collaborations", which are not very meaningful. To
# filter out loops without losing the edge weights, we can use:
g = g.simplify(combine_edges="first")

fig, ax = plt.subplots()
ig.plot(
    g,
    target=ax,
    layout=layout,
    vertex_label=g.vs["name"],
    vertex_color="lightblue",
    vertex_size=vertex_size,
    edge_width=g.es["weight"],
)
plt.show()



__all__ = (
    "_count_automorphisms_vf2",
    "_get_automorphisms_vf2",
)


def _count_automorphisms_vf2(
    graph, color=None, edge_color=None, node_compat_fn=None, edge_compat_fn=None
):
    """Returns the number of automorphisms of the graph.

    This function simply calls C{count_isomorphisms_vf2} with the graph
    itgraph. See C{count_isomorphisms_vf2} for an explanation of the
    parameters.

    @return: the number of automorphisms of the graph
    @see: Graph.count_isomorphisms_vf2
    """
    return graph.count_isomorphisms_vf2(
        graph,
        color1=color,
        color2=color,
        edge_color1=edge_color,
        edge_color2=edge_color,
        node_compat_fn=node_compat_fn,
        edge_compat_fn=edge_compat_fn,
    )


def _get_automorphisms_vf2(
    graph, color=None, edge_color=None, node_compat_fn=None, edge_compat_fn=None
):
    """Returns all the automorphisms of the graph

    This function simply calls C{get_isomorphisms_vf2} with the graph
    itgraph. See C{get_isomorphisms_vf2} for an explanation of the
    parameters.

    @return: a list of lists, each item containing a possible mapping
      of the graph vertices to itgraph according to the automorphism
    @see: Graph.get_isomorphisms_vf2
    """
    return graph.get_isomorphisms_vf2(
        graph,
        color1=color,
        color2=color,
        edge_color1=edge_color,
        edge_color2=edge_color,
        node_compat_fn=node_compat_fn,
        edge_compat_fn=edge_compat_fn,
    )



# vim:ts=4:sw=4:sts=4:et
# -*- coding: utf-8 -*-
"""Utility functions that cannot be categorised anywhere else."""

from contextlib import contextmanager

from collections.abc import MutableMapping
from itertools import chain
from warnings import warn

import os
import tempfile

__all__ = (
    "dbl_epsilon",
    "deprecated",
    "multidict",
    "named_temporary_file",
    "numpy_to_contiguous_memoryview",
    "rescale",
    "safemin",
    "safemax",
)

__docformat__ = "restructuredtext en"


def deprecated(message):
    """Prints a warning message related to the deprecation of some igraph
    feature."""
    warn(message, DeprecationWarning, stacklevel=3)


def _is_running_in_ipython():
    """Internal function that determines whether igraph is running inside
    IPython or not."""
    try:
        from IPython import get_ipython

        return get_ipython() is not None
    except ImportError:
        return False


@contextmanager
def named_temporary_file(*args, **kwds):
    """Context manager that creates a named temporary file and
    returns its name.

    All parameters are passed on to ``tempfile.mkstemp``, see
    its documentation for more info.
    """
    handle, tmpfile = tempfile.mkstemp(*args, **kwds)
    os.close(handle)
    try:
        yield tmpfile
    finally:
        os.unlink(tmpfile)


def numpy_to_contiguous_memoryview(obj):
    """Converts a NumPy array or matrix into a contiguous memoryview object
    that is suitable to be forwarded to the Graph constructor.

    This is used internally to allow us to use a NumPy array or matrix
    directly when constructing a Graph.
    """
    # Deferred import to prevent a hard dependency on NumPy
    from numpy import int32, int64, require
    from igraph._igraph import INTEGER_SIZE

    if INTEGER_SIZE == 64:
        dtype = int64
    elif INTEGER_SIZE == 32:
        dtype = int32
    else:
        raise TypeError(
            f"size of igraph_int_t in the C layer ({INTEGER_SIZE} bits) is not supported"
        )

    return memoryview(require(obj, dtype=dtype, requirements="AC"))


def rescale(values, out_range=(0.0, 1.0), in_range=None, clamp=False, scale=None):
    """Rescales a list of numbers into a given range.

    ``out_range`` gives the range of the output values; by default, the minimum
    of the original numbers in the list will be mapped to the first element
    in the output range and the maximum will be mapped to the second element.
    Elements between the minimum and maximum values in the input list will be
    interpolated linearly between the first and second values of the output
    range.

    ``in_range`` may be used to override which numbers are mapped to the first
    and second values of the output range. This must also be a tuple, where
    the first element will be mapped to the first element of the output range
    and the second element to the second.

    If ``clamp`` is ``True``, elements which are outside the given ``out_range``
    after rescaling are clamped to the output range to ensure that no number
    will be outside ``out_range`` in the result.

    If ``scale`` is not ``None``, it will be called for every element of ``values``
    and the rescaling will take place on the results instead. This can be used,
    for instance, to transform the logarithm of the original values instead of
    the actual values. A typical use-case is to map a range of values to color
    identifiers on a logarithmic scale. Scaling also applies to the ``in_range``
    parameter if present.

    :param out_range: the range of output values
    :param in_range: the range of the input values; this is the range that is mapped
        to ``out_range``. ``None`` means to use the minimum and maximum of
        the input, respectively.
    :param clamp: specifies what to do when an input value falls outside ``in_range``.
        ``True`` means to clamp the value to the bounds of ``in_range``,
        ``False`` means not to clamp.
    :param scale: an optional transformation to perform on the input values before
        mapping them to the output range.

    Examples:

        >>> rescale(range(5), (0, 8))
        [0.0, 2.0, 4.0, 6.0, 8.0]
        >>> rescale(range(5), (2, 10))
        [2.0, 4.0, 6.0, 8.0, 10.0]
        >>> rescale(range(5), (0, 4), (1, 3))
        [-2.0, 0.0, 2.0, 4.0, 6.0]
        >>> rescale(range(5), (0, 4), (1, 3), clamp=True)
        [0.0, 0.0, 2.0, 4.0, 4.0]
        >>> rescale([0]*5, (1, 3))
        [2.0, 2.0, 2.0, 2.0, 2.0]
        >>> from math import log10
        >>> rescale([1, 10, 100, 1000, 10000], (0, 8), scale=log10)
        [0.0, 2.0, 4.0, 6.0, 8.0]
        >>> rescale([1, 10, 100, 1000, 10000], (0, 4), (10, 1000), scale=log10)
        [-2.0, 0.0, 2.0, 4.0, 6.0]
    """
    if scale is not None:
        values = [scale(value) for value in values]

    if in_range is None:
        mi, ma = min(values), max(values)
    else:
        mi, ma = in_range
        if scale is not None:
            mi, ma = scale(mi), scale(ma)

    ratio = float(ma - mi)
    if not ratio:
        return [(out_range[0] + out_range[1]) / 2.0] * len(values)

    min_out, max_out = list(map(float, out_range))
    ratio = (max_out - min_out) / ratio
    result = [(x - mi) * ratio + min_out for x in values]

    if clamp:
        return [max(min(x, max_out), min_out) for x in result]
    else:
        return result


def consecutive_pairs(iterable, circular=False):
    """Returns consecutive pairs of items from the given iterable.

    When ``circular`` is ``True``, the pair consisting of the last
    and first elements is also returned.

    Example:

        >>> list(consecutive_pairs(range(5)))
        [(0, 1), (1, 2), (2, 3), (3, 4)]
        >>> list(consecutive_pairs(range(5), circular=True))
        [(0, 1), (1, 2), (2, 3), (3, 4), (4, 0)]
        >>> list(consecutive_pairs([]))
        []
        >>> list(consecutive_pairs([], circular=True))
        []
        >>> list(consecutive_pairs([0]))
        []
        >>> list(consecutive_pairs([0], circular=True))
        [(0, 0)]
    """
    it = iter(iterable)

    try:
        prev = next(it)
    except StopIteration:
        return
    first = prev

    for item in it:
        yield prev, item
        prev = item

    if circular:
        try:
            yield item, first
        except UnboundLocalError:
            yield first, first


class multidict(MutableMapping):
    """A dictionary-like object that is customized to deal with multiple
    values for the same key.

    Each value in this dictionary will be a list. Methods which emulate
    the methods of a standard Python ``dict`` object will return or manipulate
    the first items of the lists only. Special methods are provided to
    deal with keys having multiple values.
    """

    def __init__(self, *args, **kwds):
        self._dict = {}
        if len(args) > 1:
            raise ValueError(
                "%r expected at most 1 argument, got %d"
                % (self.__class__.__name__, len(args))
            )
        if args:
            args = args[0]
            self.update(args)
        self.update(kwds)

    def __contains__(self, key):
        """Returns whether there are any items associated to the given
        ``key``."""
        try:
            return len(self._dict[key]) > 0
        except KeyError:
            return False

    def __delitem__(self, key):
        """Removes all the items associated to the given ``key``."""
        del self._dict[key]

    def __getitem__(self, key):
        """Returns an arbitrary item associated to the given key. Raises ``KeyError``
        if no such key exists.

        Example:

            >>> d = multidict([("spam", "eggs"), ("spam", "bacon")])
            >>> d["spam"]
            'eggs'
        """
        try:
            return self._dict[key][0]
        except IndexError:
            raise KeyError(key) from None

    def __iter__(self):
        """Iterates over the keys of the multidict."""
        return iter(self._dict)

    def __len__(self):
        """Returns the number of distinct keys in this multidict."""
        return len(self._dict)

    def __setitem__(self, key, value):
        """Sets the item associated to the given ``key``. Any values associated to the
        key will be erased and replaced by ``value``.

        Example:

           >>> d = multidict([("spam", "eggs"), ("spam", "bacon")])
           >>> d["spam"] = "ham"
           >>> d["spam"]
           'ham'
        """
        self._dict[key] = [value]

    def add(self, key, value):
        """Adds `value` to the list of items associated to ``key``.

        Example:

            >>> d = multidict()
            >>> d.add("spam", "ham")
            >>> d["spam"]
            'ham'
            >>> d.add("spam", "eggs")
            >>> d.getlist("spam")
            ['ham', 'eggs']
        """
        try:
            self._dict[key].append(value)
        except KeyError:
            self._dict[key] = [value]

    def clear(self):
        """Removes all the items from the multidict."""
        self._dict.clear()

    def get(self, key, default=None):
        """Returns an arbitrary item associated to the given ``key``. If ``key``
        does not exist or has zero associated items, ``default`` will be
        returned."""
        try:
            items = self._dict[key]
            return items[0]
        except (KeyError, IndexError):
            return default

    def getlist(self, key):
        """Returns the list of values for the given ``key``. An empty list will
        be returned if there is no such key."""
        try:
            return self._dict[key]
        except KeyError:
            return []

    def iterlists(self):
        """Iterates over ``(key, values)`` pairs where ``values`` is the list
        of values associated with ``key``."""
        return iter(self._dict.items())

    def lists(self):
        """Returns a list of ``(key, values)`` pairs where ``values`` is the list
        of values associated with ``key``."""
        return list(self._dict.items())

    def update(self, arg, **kwds):
        if hasattr(arg, "keys") and callable(arg.keys):
            for key in arg.keys():
                self.add(key, arg[key])
        else:
            for key, value in arg:
                self.add(key, value)
        for key, value in kwds.items():
            self.add(key, value)


def safemax(iterable, default=0):
    """Safer variant of ``max()`` that returns a default value if the iterable
    is empty.

    Example:

        >>> safemax([-5, 6, 4])
        6
        >>> safemax([])
        0
        >>> safemax((), 2)
        2
    """
    it = iter(iterable)
    try:
        first = next(it)
    except StopIteration:
        return default
    else:
        return max(chain([first], it))


def safemin(iterable, default=0):
    """Safer variant of ``min()`` that returns a default value if the iterable
    is empty.

    Example:

        >>> safemin([-5, 6, 4])
        -5
        >>> safemin([])
        0
        >>> safemin((), 2)
        2
    """
    it = iter(iterable)
    try:
        first = next(it)
    except StopIteration:
        return default
    else:
        return min(chain([first], it))


def dbl_epsilon():
    """Approximates the machine epsilon value for doubles."""
    epsilon = 1.0
    while 1.0 + epsilon / 2.0 != 1.0:
        epsilon /= 2
    return epsilon


dbl_epsilon = dbl_epsilon()



"""This module provides implementations of Cairo-specific vertex drawers, i.e.
drawers that the Cairo graph drawer will use to draw vertices.
"""

from math import pi

from igraph.drawing.baseclasses import AbstractVertexDrawer
from igraph.drawing.metamagic import AttributeCollectorBase
from igraph.drawing.shapes import ShapeDrawerDirectory

from .base import AbstractCairoDrawer

__all__ = ("AbstractCairoVertexDrawer", "CairoVertexDrawer")


class AbstractCairoVertexDrawer(AbstractVertexDrawer, AbstractCairoDrawer):
    """Abstract base class for vertex drawers that draw on a Cairo canvas."""

    def __init__(self, context, bbox, palette, layout):
        """Constructs the vertex drawer and associates it to the given
        Cairo context and the given L{BoundingBox}.

        @param context: the context on which we will draw
        @param bbox:    the bounding box within which we will draw.
                        Can be anything accepted by the constructor
                        of L{BoundingBox} (i.e., a 2-tuple, a 4-tuple
                        or a L{BoundingBox} object).
        @param palette: the palette that can be used to map integer
                        color indices to colors when drawing vertices
        @param layout:  the layout of the vertices in the graph being drawn
        """
        AbstractCairoDrawer.__init__(self, context, bbox)
        AbstractVertexDrawer.__init__(self, palette, layout)


class CairoVertexDrawer(AbstractCairoVertexDrawer):
    """The default vertex drawer implementation of igraph."""

    def __init__(self, context, bbox, palette, layout):
        super().__init__(context, bbox, palette, layout)
        self.VisualVertexBuilder = self._construct_visual_vertex_builder()

    def _construct_visual_vertex_builder(self):
        class VisualVertexBuilder(AttributeCollectorBase):
            """Collects some visual properties of a vertex for drawing"""

            _kwds_prefix = "vertex_"
            color = ("red", self.palette.get)
            frame_color = ("black", self.palette.get)
            frame_width = 1.0
            label = None
            label_angle = -pi / 2
            label_dist = 0.0
            label_color = ("black", self.palette.get)
            font = "sans-serif"
            label_size = 14.0
            position = {"func": self.layout.__getitem__}
            shape = ("circle", ShapeDrawerDirectory.resolve_default)
            size = 20.0
            width = None
            height = None

        return VisualVertexBuilder

    def draw(self, visual_vertex, vertex, coords):
        context = self.context

        width = (
            visual_vertex.width
            if visual_vertex.width is not None
            else visual_vertex.size
        )
        height = (
            visual_vertex.height
            if visual_vertex.height is not None
            else visual_vertex.size
        )

        visual_vertex.shape.draw_path(context, coords[0], coords[1], width, height)
        context.set_source_rgba(*visual_vertex.color)
        context.fill_preserve()
        context.set_source_rgba(*visual_vertex.frame_color)
        context.set_line_width(visual_vertex.frame_width)
        context.stroke()



import math

from igraph.drawing import BoundingBox


def _write_graph_to_svg(
    graph,
    fname,
    layout="auto",
    width=None,
    height=None,
    labels="label",
    colors="color",
    shapes="shape",
    vertex_size=10,
    edge_colors="color",
    edge_stroke_widths="width",
    font_size=16,
    *args,
    **kwds,
):
    """Saves the graph as an SVG (Scalable Vector Graphics) file

    The file will be Inkscape (http://inkscape.org) compatible.
    In Inkscape, as nodes are rearranged, the edges auto-update.

    @param fname: the name of the file or a Python file handle
    @param layout: the layout of the graph. Can be either an
      explicitly specified layout (using a list of coordinate
      pairs) or the name of a layout algorithm (which should
      refer to a method in the L{Graph} object, but without
      the C{layout_} prefix.
    @param width: the preferred width in pixels (default: 400)
    @param height: the preferred height in pixels (default: 400)
    @param labels: the vertex labels. Either it is the name of
      a vertex attribute to use, or a list explicitly specifying
      the labels. It can also be C{None}.
    @param colors: the vertex colors. Either it is the name of
      a vertex attribute to use, or a list explicitly specifying
      the colors. A color can be anything acceptable in an SVG
      file.
    @param shapes: the vertex shapes. Either it is the name of
      a vertex attribute to use, or a list explicitly specifying
      the shapes as integers. Shape 0 means hidden (nothing is drawn),
      shape 1 is a circle, shape 2 is a rectangle and shape 3 is a
      rectangle that automatically sizes to the inner text.
    @param vertex_size: vertex size in pixels
    @param edge_colors: the edge colors. Either it is the name
      of an edge attribute to use, or a list explicitly specifying
      the colors. A color can be anything acceptable in an SVG
      file.
    @param edge_stroke_widths: the stroke widths of the edges. Either
      it is the name of an edge attribute to use, or a list explicitly
      specifying the stroke widths. The stroke width can be anything
      acceptable in an SVG file.
    @param font_size: font size. If it is a string, it is written into
      the SVG file as-is (so you can specify anything which is valid
      as the value of the C{font-size} style). If it is a number, it
      is interpreted as pixel size and converted to the proper attribute
      value accordingly.
    """
    if width is None and height is None:
        width = 400
        height = 400
    elif width is None:
        width = height
    elif height is None:
        height = width

    if width <= 0 or height <= 0:
        raise ValueError("width and height must be positive")

    if isinstance(layout, str):
        layout = graph.layout(layout, *args, **kwds)

    if isinstance(labels, str):
        try:
            labels = graph.vs.get_attribute_values(labels)
        except KeyError:
            labels = [x + 1 for x in range(graph.vcount())]
    elif labels is None:
        labels = [""] * graph.vcount()

    if isinstance(colors, str):
        try:
            colors = graph.vs.get_attribute_values(colors)
        except KeyError:
            colors = ["red"] * graph.vcount()

    if isinstance(shapes, str):
        try:
            shapes = graph.vs.get_attribute_values(shapes)
        except KeyError:
            shapes = [1] * graph.vcount()

    if isinstance(edge_colors, str):
        try:
            edge_colors = graph.es.get_attribute_values(edge_colors)
        except KeyError:
            edge_colors = ["black"] * graph.ecount()

    if isinstance(edge_stroke_widths, str):
        try:
            edge_stroke_widths = graph.es.get_attribute_values(edge_stroke_widths)
        except KeyError:
            edge_stroke_widths = [2] * graph.ecount()

    if not isinstance(font_size, str):
        font_size = "%spx" % str(font_size)
    else:
        if ";" in font_size:
            raise ValueError("font size can't contain a semicolon")

    vcount = graph.vcount()
    labels.extend(str(i + 1) for i in range(len(labels), vcount))
    colors.extend(["red"] * (vcount - len(colors)))

    if isinstance(fname, str):
        f = open(fname, "w")
        our_file = True
    else:
        f = fname
        our_file = False

    bbox = BoundingBox(layout.bounding_box())

    sizes = [width - 2 * vertex_size, height - 2 * vertex_size]
    w, h = bbox.width, bbox.height

    ratios = []
    if w == 0:
        ratios.append(1.0)
    else:
        ratios.append(sizes[0] / w)
    if h == 0:
        ratios.append(1.0)
    else:
        ratios.append(sizes[1] / h)

    layout = [
        [
            (row[0] - bbox.left) * ratios[0] + vertex_size,
            (row[1] - bbox.top) * ratios[1] + vertex_size,
        ]
        for row in layout
    ]

    directed = graph.is_directed()

    print('<?xml version="1.0" encoding="UTF-8" standalone="no"?>', file=f)
    print(
        "<!-- Created by igraph (https://igraph.org/) -->",
        file=f,
    )
    print(file=f)
    print(
        '<svg xmlns:dc="http://purl.org/dc/elements/1.1/" '
        'xmlns:cc="http://creativecommons.org/ns#" '
        'xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" '
        'xmlns:svg="http://www.w3.org/2000/svg" '
        'xmlns="http://www.w3.org/2000/svg" '
        'xmlns:sodipodi="http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd" '
        'xmlns:inkscape="http://www.inkscape.org/namespaces/inkscape"',
        file=f,
    )
    print('width="{0}px" height="{1}px">'.format(width, height), end=" ", file=f)

    edge_color_dict = {}
    print('<defs id="defs3">', file=f)
    for e_col in set(edge_colors):
        if e_col == "#000000":
            marker_index = ""
        else:
            marker_index = str(len(edge_color_dict))
        # Print an arrow marker for each possible line color
        # This is a copy of Inkscape's standard Arrow 2 marker
        print("<marker", file=f)
        print('   inkscape:stockid="Arrow2Lend{0}"'.format(marker_index), file=f)
        print('   orient="auto"', file=f)
        print('   refY="0.0"', file=f)
        print('   refX="0.0"', file=f)
        print('   id="Arrow2Lend{0}"'.format(marker_index), file=f)
        print('   style="overflow:visible;">', file=f)
        print("  <path", file=f)
        print('     id="pathArrow{0}"'.format(marker_index), file=f)
        print(
            '     style="font-size:12.0;fill-rule:evenodd;'
            "stroke-width:0.62500000;stroke-linejoin:round;"
            'fill:{0}"'.format(e_col),
            file=f,
        )
        print(
            '     d="M 8.7185878,4.0337352 L -2.2072895,0.016013256 '
            "L 8.7185884,-4.0017078 C 6.9730900,-1.6296469 "
            '6.9831476,1.6157441 8.7185878,4.0337352 z "',
            file=f,
        )
        print('     transform="scale(1.1) rotate(180) translate(1,0)" />', file=f)
        print("</marker>", file=f)

        edge_color_dict[e_col] = "Arrow2Lend{0}".format(marker_index)
    print("</defs>", file=f)
    print(
        '<g inkscape:groupmode="layer" id="layer2" inkscape:label="Lines" '
        'sodipodi:insensitive="true">',
        file=f,
    )

    for eidx, edge in enumerate(graph.es):
        vidxs = edge.tuple
        x1 = layout[vidxs[0]][0]
        y1 = layout[vidxs[0]][1]
        x2 = layout[vidxs[1]][0]
        y2 = layout[vidxs[1]][1]
        angle = math.atan2(y2 - y1, x2 - x1)
        x2 -= vertex_size * math.cos(angle)
        y2 -= vertex_size * math.sin(angle)

        print("<path", file=f)
        print(
            '    style="fill:none;stroke:{0};stroke-width:{2};'
            "stroke-linecap:butt;stroke-linejoin:miter;stroke-miterlimit:4;"
            'stroke-opacity:1;stroke-dasharray:none{1}"'.format(
                edge_colors[eidx],
                ";marker-end:url(#{0})".format(edge_color_dict[edge_colors[eidx]])
                if directed
                else "",
                edge_stroke_widths[eidx],
            ),
            file=f,
        )
        print('    d="M {0},{1} {2},{3}"'.format(x1, y1, x2, y2), file=f)
        print('    id="path{0}"'.format(eidx), file=f)
        print('    inkscape:connector-type="polyline"', file=f)
        print('    inkscape:connector-curvature="0"', file=f)
        print('    inkscape:connection-start="#g{0}"'.format(edge.source), file=f)
        print('    inkscape:connection-start-point="d4"', file=f)
        print('    inkscape:connection-end="#g{0}"'.format(edge.target), file=f)
        print('    inkscape:connection-end-point="d4" />', file=f)

    print("  </g>", file=f)
    print(file=f)

    print(
        '  <g inkscape:label="Nodes" \
                    inkscape:groupmode="layer" id="layer1">',
        file=f,
    )
    print("  <!-- Vertices -->", file=f)

    if any(x == 3 for x in shapes):
        # Only import tkFont if we really need it. Unfortunately, this will
        # flash up an unneccesary Tk window in some cases
        import tkinter.font
        import tkinter as tk

        # This allows us to dynamically size the width of the nodes.
        # Unfortunately this works only with font sizes specified in pixels.
        if font_size.endswith("px"):
            font_size_in_pixels = int(font_size[:-2])
        else:
            try:
                font_size_in_pixels = int(font_size)
            except Exception:
                raise ValueError(
                    "font sizes must be specified in pixels "
                    "when any of the nodes has shape=3 (i.e. "
                    "node size determined by text size)"
                ) from None
        tk_window = tk.Tk()
        font = tkinter.font.Font(
            root=tk_window, font=("Sans", font_size_in_pixels, tkinter.font.NORMAL)
        )
    else:
        tk_window = None

    for vidx in range(graph.vcount()):
        print(
            '    <g id="g{0}" transform="translate({1},{2})">'.format(
                vidx, layout[vidx][0], layout[vidx][1]
            ),
            file=f,
        )
        if shapes[vidx] == 1:
            # Undocumented feature: can handle two colors but only for circles
            c = str(colors[vidx])
            if " " in c:
                c = c.split(" ")
                vs = str(vertex_size)
                print(
                    '     <path d="M -{0},0 A{0},{0} 0 0,0 {0},0 L \
                                -{0},0" fill="{1}"/>'.format(vs, c[0]),
                    file=f,
                )
                print(
                    '     <path d="M -{0},0 A{0},{0} 0 0,1 {0},0 L \
                                -{0},0" fill="{1}"/>'.format(vs, c[1]),
                    file=f,
                )
                print(
                    '     <circle cx="0" cy="0" r="{0}" fill="none"/>'.format(vs),
                    file=f,
                )
            else:
                print(
                    '     <circle cx="0" cy="0" r="{0}" fill="{1}"/>'.format(
                        str(vertex_size), str(colors[vidx])
                    ),
                    file=f,
                )
        elif shapes[vidx] == 2:
            print(
                '      <rect x="-{0}" y="-{0}" width="{1}" height="{1}" '
                'id="rect{2}" style="fill:{3};fill-opacity:1" />'.format(
                    vertex_size, vertex_size * 2, vidx, colors[vidx]
                ),
                file=f,
            )
        elif shapes[vidx] == 3:
            (vertex_width, vertex_height) = (
                font.measure(str(labels[vidx])) + 2,
                font.metrics("linespace") + 2,
            )
            print(
                '      <rect ry="5" rx="5" x="-{0}" y="-{1}" width="{2}" '
                'height="{3}" id="rect{4}" style="fill:{5};fill-opacity:1" '
                "/>".format(
                    vertex_width / 2.0,
                    vertex_height / 2.0,
                    vertex_width,
                    vertex_height,
                    vidx,
                    colors[vidx],
                ),
                file=f,
            )

        print(
            '      <text sodipodi:linespacing="125%" y="{0}" x="0" '
            'id="text{1}" style="font-size:{2};font-style:normal;'
            "font-weight:normal;text-align:center;line-height:125%;"
            "letter-spacing:0px;word-spacing:0px;text-anchor:middle;"
            "fill:#000000;fill-opacity:1;stroke:none;"
            'font-family:Sans">'.format(vertex_size / 2.0, vidx, font_size),
            file=f,
        )
        print(
            '<tspan y="{0}" x="0" id="tspan{1}" sodipodi:role="line">'
            "{2}</tspan></text>".format(vertex_size / 2.0, vidx, str(labels[vidx])),
            file=f,
        )
        print("    </g>", file=f)

    print("</g>", file=f)
    print(file=f)
    print("</svg>", file=f)

    if our_file:
        f.close()
    if tk_window:
        tk_window.destroy()



"""
.. _tutorials-connected-components:

=====================
Connected Components
=====================

This example demonstrates how to visualise the connected components in a graph using :meth:`igraph.GraphBase.connected_components`.
"""

import igraph as ig
import matplotlib.pyplot as plt
import random

# %%
# First, we generate a randomized geometric graph with random vertex sizes. The
# seed is set to the example is reproducible in our manual: you don't really
# need it to understand the concepts.
random.seed(0)
g = ig.Graph.GRG(50, 0.15)

# %%
# Now we can cluster the graph into weakly connected components, i.e. subgraphs
# that have no edges connecting them to one another:
components = g.connected_components(mode="weak")

# %%
# Finally, we can visualize the distinct connected components of the graph:
fig, ax = plt.subplots()
ig.plot(
    components,
    target=ax,
    palette=ig.RainbowPalette(),
    vertex_size=7,
    vertex_color=list(map(int, ig.rescale(components.membership, (0, 200), clamp=True))),
    edge_width=0.7
)
plt.show()

# %%
# .. note::
#
#     We use the integers from 0 to 200 instead of 0 to 255 in our vertex
#     colors, since 255 in the :class:`igraph.drawing.colors.RainbowPalette`
#     corresponds to looping back to red. This gives us nicely distinct hues.



"""
.. _tutorials-betweenness:

=======================
Betweenness
=======================

This example demonstrates how to visualize both vertex and edge betweenness with a custom defined color palette. We use the methods :meth:`igraph.GraphBase.betweenness` and :meth:`igraph.GraphBase.edge_betweenness` respectively, and demonstrate the effects on a standard `Krackhardt Kite <https://www.wikiwand.com/en/Krackhardt_kite_graph>`_ graph, as well as a `Watts-Strogatz <https://en.wikipedia.org/wiki/Watts%E2%80%93Strogatz_model>`_ random graph.

"""
import random
import matplotlib.pyplot as plt
from matplotlib.cm import ScalarMappable
from matplotlib.colors import LinearSegmentedColormap, Normalize
import igraph as ig


# %%
# We define a function that plots the graph  on a Matplotlib axis, along with
# its vertex and edge betweenness values. The function also generates some
# color bars on the sides to see how they translate to each other. We use
# `Matplotlib's Normalize class <https://matplotlib.org/stable/api/_as_gen/matplotlib.colors.Normalize.html>`_
# to ensure that our color bar ranges are correct, as well as *igraph*'s
# :meth:`igraph.utils.rescale` to rescale the betweennesses in the interval
# ``[0, 1]``.
def plot_betweenness(g, vertex_betweenness, edge_betweenness, ax, cax1, cax2):
    """Plot vertex/edge betweenness, with colorbars

    Args:
        g: the graph to plot.
        ax: the Axes for the graph
        cax1: the Axes for the vertex betweenness colorbar
        cax2: the Axes for the edge betweenness colorbar
    """

    # Rescale betweenness to be between 0.0 and 1.0
    scaled_vertex_betweenness = ig.rescale(vertex_betweenness, clamp=True)
    scaled_edge_betweenness = ig.rescale(edge_betweenness, clamp=True)
    print(f"vertices: {min(vertex_betweenness)} - {max(vertex_betweenness)}")
    print(f"edges: {min(edge_betweenness)} - {max(edge_betweenness)}")

    # Define mappings betweenness -> color
    cmap1 = LinearSegmentedColormap.from_list("vertex_cmap", ["pink", "indigo"])
    cmap2 = LinearSegmentedColormap.from_list("edge_cmap", ["lightblue", "midnightblue"])

    # Plot graph
    g.vs["color"] = [cmap1(betweenness) for betweenness in scaled_vertex_betweenness]
    g.vs["size"] = ig.rescale(vertex_betweenness, (10, 50))
    g.es["color"] = [cmap2(betweenness) for betweenness in scaled_edge_betweenness]
    g.es["width"] = ig.rescale(edge_betweenness, (0.5, 1.0))
    ig.plot(
        g,
        target=ax,
        layout="fruchterman_reingold",
        vertex_frame_width=0.2,
    )

    # Color bars
    norm1 = ScalarMappable(norm=Normalize(0, max(vertex_betweenness)), cmap=cmap1)
    norm2 = ScalarMappable(norm=Normalize(0, max(edge_betweenness)), cmap=cmap2)
    plt.colorbar(norm1, cax=cax1, orientation="horizontal", label="Vertex Betweenness")
    plt.colorbar(norm2, cax=cax2, orientation="horizontal", label="Edge Betweenness")


# %%
# First, generate a graph, e.g. the Krackhardt Kite Graph:
random.seed(0)
g1 = ig.Graph.Famous("Krackhardt_Kite")

# %%
# Then we can compute vertex and edge betweenness:
vertex_betweenness1 = g1.betweenness()
edge_betweenness1 = g1.edge_betweenness()

# %% As a second example, we generate and analyze a Watts Strogatz graph:
g2 = ig.Graph.Watts_Strogatz(dim=1, size=150, nei=2, p=0.1)
vertex_betweenness2 = g2.betweenness()
edge_betweenness2 = g2.edge_betweenness()

# %%
# Finally, we plot the two graphs, each with two colorbars for vertex/edge
# betweenness
fig, axs = plt.subplots(
    3, 2,
    figsize=(7, 6),
    gridspec_kw={"height_ratios": (20, 1, 1)},
)
plot_betweenness(g1, vertex_betweenness1, edge_betweenness1, *axs[:, 0])
plot_betweenness(g2, vertex_betweenness2, edge_betweenness2, *axs[:, 1])
fig.tight_layout(h_pad=1)
plt.show()



"""
.. _tutorials-shortest-paths:

==============
Shortest Paths
==============

This example demonstrates how to find the shortest distance between two vertices
of a weighted or an unweighted graph.
"""

import igraph as ig
import matplotlib.pyplot as plt

# %%
# To find the shortest path or distance between two nodes, we can use :meth:`igraph.GraphBase.get_shortest_paths`. If we're only interested in counting the unweighted distance, then we can do the following:
g = ig.Graph(6, [(0, 1), (0, 2), (1, 3), (2, 3), (2, 4), (3, 5), (4, 5)])
results = g.get_shortest_paths(1, to=4, output="vpath")

# results = [[1, 0, 2, 4]]

# %%
# We can print the result of the computation:
if len(results[0]) > 0:
    # The distance is the number of vertices in the shortest path minus one.
    print("Shortest distance is: ", len(results[0]) - 1)
else:
    print("End node could not be reached!")

# %%
# If the edges have weights, things are a little different. First, let's add
# weights to our graph edges:
g.es["weight"] = [2, 1, 5, 4, 7, 3, 2]

# %%
# To get the shortest paths on a weighted graph, we pass the weights as an
# argument. For a change, we choose the output format as ``"epath"`` to
# receive the path as an edge list, which can be used to calculate the length
# of the path.
results = g.get_shortest_paths(0, to=5, weights=g.es["weight"], output="epath")

# results = [[1, 3, 5]]

if len(results[0]) > 0:
    # Add up the weights across all edges on the shortest path
    distance = 0
    for e in results[0]:
        distance += g.es[e]["weight"]
    print("Shortest weighted distance is: ", distance)
else:
    print("End node could not be reached!")

# %%
# .. note::
#
#     - :meth:`igraph.GraphBase.get_shortest_paths` returns a list of lists becuase the `to` argument can also accept a list of vertex IDs. In that case, the shortest path to all each vertex is found and stored in the results array.
#     - If you're interested in finding *all* shortest paths, take a look at :meth:`igraph.GraphBase.get_all_shortest_paths`.

# %%
# In case you are wondering how the visualization figure was done, here's the code:
g.es["width"] = 0.5
g.es[results[0]]["width"] = 2.5

fig, ax = plt.subplots()
ig.plot(
    g,
    target=ax,
    layout="circle",
    vertex_color="steelblue",
    vertex_label=range(g.vcount()),
    edge_width=g.es["width"],
    edge_label=g.es["weight"],
    edge_color="#666",
    edge_align_label=True,
    edge_background="white",
)
plt.show()



# vim:ts=4:sw=4:sts=4:et
# -*- coding: utf-8 -*-
"""Implementation of union, disjoint union and intersection operators."""

__all__ = (
    "disjoint_union",
    "union",
    "intersection",
    "operator_method_registry",
)

from igraph.operators.functions import (
    disjoint_union,
    union,
    intersection,
)
from igraph.operators.methods import (
    __iadd__,
    __add__,
    __and__,
    __isub__,
    __sub__,
    __mul__,
    __or__,
    _disjoint_union,
    _union,
    _intersection,
)

operator_method_registry = {
    "__iadd__": __iadd__,
    "__add__": __add__,
    "__and__": __and__,
    "__isub__": __isub__,
    "__sub__": __sub__,
    "__mul__": __mul__,
    "__or__": __or__,
    "disjoint_union": _disjoint_union,
    "union": _union,
    "intersection": _intersection,
}



# vim:ts=4:sw=4:sts=4:et
# -*- coding: utf-8 -*-
"""Summary representation of a graph."""

import sys

from igraph.statistics import median
from itertools import islice
from math import ceil
from texttable import Texttable
from textwrap import TextWrapper

__all__ = ("GraphSummary", "summary")


class FakeWrapper:
    """Object whose interface is compatible with C{textwrap.TextWrapper}
    but does no wrapping."""

    def __init__(self, *args, **kwds):
        pass

    def fill(self, text):
        return [text]

    def wrap(self, text):
        return [text]


def _get_wrapper_for_width(width, *args, **kwds):
    """Returns a text wrapper that wraps text for the given width.

    @param width: the maximal width of each line that the text wrapper
      produces. C{None} means that no wrapping will be performed.
    """
    if width is None:
        return FakeWrapper(*args, **kwds)
    return TextWrapper(width, *args, **kwds)


class GraphSummary:
    """Summary representation of a graph.

    The summary representation includes a header line and the list of
    edges. The header line consists of C{IGRAPH}, followed by a
    four-character long code, the number of vertices, the number of
    edges, two dashes (C{--}) and the name of the graph (i.e.
    the contents of the C{name} attribute, if any). For instance,
    a header line may look like this::

        IGRAPH U--- 4 5 --

    The four-character code describes some basic properties of the
    graph. The first character is C{U} if the graph is undirected,
    C{D} if it is directed. The second letter is C{N} if the graph
    has a vertex attribute called C{name}, or a dash otherwise. The
    third letter is C{W} if the graph is weighted (i.e. it has an
    edge attribute called C{weight}), or a dash otherwise. The
    fourth letter is C{B} if the graph has a vertex attribute called
    C{type}; this is usually used for bipartite graphs.

    Edges may be presented as an ordinary edge list or an adjacency
    list. By default, this depends on the number of edges; however,
    you can control it with the appropriate constructor arguments.
    """

    def __init__(
        self,
        graph,
        verbosity=0,
        width=78,
        edge_list_format="auto",
        max_rows=99999,
        print_graph_attributes=False,
        print_vertex_attributes=False,
        print_edge_attributes=False,
        full=False,
    ):
        """Constructs a summary representation of a graph.

        @param verbosity: the verbosity of the summary. If zero, only
          the header line will be returned. If one, the header line
          and the list of edges will both be returned.
        @param width: the maximal width of each line in the summary.
          C{None} means that no limit will be enforced.
        @param max_rows: the maximal number of rows to print in a single
          table (e.g., vertex attribute table or edge attribute table)
        @param edge_list_format: format of the edge list in the summary.
          Supported formats are: C{compressed}, C{adjlist}, C{edgelist},
          C{auto}, which selects automatically from the other three based
          on some simple criteria.
        @param print_graph_attributes: whether to print graph attributes
          if there are any.
        @param print_vertex_attributes: whether to print vertex attributes
          if there are any.
        @param print_edge_attributes: whether to print edge attributes
          if there are any.
        @param full: False has no effect; True turns on the attribute
          printing for graph, vertex and edge attributes with verbosity 1.
        """
        if full:
            print_graph_attributes = True
            print_vertex_attributes = True
            print_edge_attributes = True
            verbosity = max(verbosity, 1)

        self._graph = graph
        self.edge_list_format = edge_list_format.lower()
        self.max_rows = int(max_rows)
        self.print_graph_attributes = print_graph_attributes
        self.print_vertex_attributes = print_vertex_attributes
        self.print_edge_attributes = print_edge_attributes
        self.verbosity = verbosity
        self.width = width
        self.wrapper = _get_wrapper_for_width(self.width, break_on_hyphens=False)

        if self._graph.is_named():
            self._edges_header = "+ edges (vertex names):"
        else:
            self._edges_header = "+ edges:"
        self._arrow = ["--", "->"][self._graph.is_directed()]
        self._arrow_format = "%%s%s%%s" % self._arrow

    def _construct_edgelist_adjlist(self):
        """Constructs the part in the summary that prints the edge list in an
        adjacency list format."""
        result = [self._edges_header]

        if self._graph.vcount() == 0:
            return

        if self._graph.is_named():
            names = self._graph.vs["name"]
            maxlen = max(len(str(name)) for name in names)
            format_str = "%%%ds %s %%s" % (maxlen, self._arrow)
            for v1, name in enumerate(names):
                neis = self._graph.successors(v1)
                neis = ", ".join(str(names[v2]) for v2 in neis)
                result.append(format_str % (name, neis))
        else:
            maxlen = len(str(self._graph.vcount()))
            num_format = "%%%dd" % maxlen
            format_str = "%s %s %%s" % (num_format, self._arrow)
            for v1 in range(self._graph.vcount()):
                neis = self._graph.successors(v1)
                neis = " ".join(num_format % v2 for v2 in neis)
                result.append(format_str % (v1, neis))

        # Try to wrap into multiple columns if that works with the given width
        if self.width is not None:
            maxlen = max(len(line) for line in result[1:])
            colcount = int(self.width + 3) / int(maxlen + 3)
            if colcount > 1:
                # Rewrap to multiple columns
                nrows = len(result) - 1
                colheight = int(ceil(nrows / float(colcount)))
                newrows = [[] for _ in range(colheight)]
                for i, row in enumerate(result[1:]):
                    newrows[i % colheight].append(row.ljust(maxlen))
                result[1:] = ["   ".join(row) for row in newrows]

        return result

    def _construct_edgelist_compressed(self):
        """Constructs the part in the summary that prints the edge list in a
        compressed format suitable for graphs with mostly small degrees."""
        result = [self._edges_header]
        arrow = self._arrow_format

        if self._graph.is_named():
            names = self._graph.vs["name"]
            edges = ", ".join(
                arrow % (names[edge.source], names[edge.target])
                for edge in self._graph.es
            )
        else:
            edges = " ".join(arrow % edge.tuple for edge in self._graph.es)

        result.append(edges)
        return result

    def _construct_edgelist_edgelist(self):
        """Constructs the part in the summary that prints the edge list in a
        full edge list format."""
        attrs = sorted(self._graph.edge_attributes())

        table = self._new_table(headers=["", "edge"] + attrs)
        table.add_rows(
            islice(self._edge_attribute_iterator(attrs), 0, self.max_rows), header=False
        )
        table.set_cols_align(
            ["l", "l"] + self._infer_column_alignment(edge_attrs=attrs)
        )

        result = [self._edges_header]
        result.extend(table.draw().split("\n"))

        return result

    def _construct_graph_attributes(self):
        """Constructs the part in the summary that lists the graph attributes."""
        attrs = self._graph.attributes()
        if not attrs:
            return []

        result = ["+ graph attributes:"]
        attrs.sort()
        for attr in attrs:
            result.append("[[%s]]" % (attr,))
            result.append(str(self._graph[attr]))
        return result

    def _construct_vertex_attributes(self):
        """Constructs the part in the summary that lists the vertex attributes."""
        attrs = sorted(self._graph.vertex_attributes())
        if not attrs or (len(attrs) == 1 and "name" in attrs):
            return []

        table = self._new_table(headers=[""] + attrs)
        table.add_rows(
            islice(self._vertex_attribute_iterator(attrs), 0, self.max_rows),
            header=False,
        )
        table.set_cols_align(["l"] + self._infer_column_alignment(vertex_attrs=attrs))

        result = ["+ vertex attributes:"]
        result.extend(table.draw().split("\n"))

        return result

    def _construct_header(self):
        """Constructs the header part of the summary."""
        graph = self._graph
        params = {
            "directed": "UD"[graph.is_directed()],
            "named": "-N"[graph.is_named()],
            "weighted": "-W"[graph.is_weighted()],
            "typed": "-T"["type" in graph.vertex_attributes()],
            "vcount": graph.vcount(),
            "ecount": graph.ecount(),
        }
        if "name" in graph.attributes():
            params["name"] = graph["name"]
        else:
            params["name"] = ""
        result = [
            "IGRAPH %(directed)s%(named)s%(weighted)s%(typed)s "
            "%(vcount)d %(ecount)d -- %(name)s" % params
        ]

        attrs = ["%s (g)" % (name,) for name in sorted(graph.attributes())]
        attrs.extend("%s (v)" % (name,) for name in sorted(graph.vertex_attributes()))
        attrs.extend("%s (e)" % (name,) for name in sorted(graph.edge_attributes()))
        if attrs:
            result.append("+ attr: %s" % ", ".join(attrs))
            if self.wrapper is not None:
                self.wrapper.subsequent_indent = "  "
                result[-1:] = self.wrapper.wrap(result[-1])
                self.wrapper.subsequent_indent = ""

        return result

    def _edge_attribute_iterator(self, attribute_order):
        """Returns an iterator that yields the rows of the edge attribute table
        in the summary. C{attribute_order} must be a list containing the names of
        the attributes to be presented in this table."""
        arrow = self._arrow_format

        if self._graph.is_named():
            names = self._graph.vs["name"]
            for edge in self._graph.es:
                formatted_edge = arrow % (names[edge.source], names[edge.target])
                yield ["[%d]" % edge.index, formatted_edge] + [
                    edge[attr] for attr in attribute_order
                ]
        else:
            for edge in self._graph.es:
                formatted_edge = arrow % edge.tuple
                yield ["[%d]" % edge.index, formatted_edge] + [
                    edge[attr] for attr in attribute_order
                ]

    def _infer_column_alignment(self, vertex_attrs=None, edge_attrs=None):
        """Infers the preferred alignment for the given vertex and edge attributes
        in the tables by peeking into the attribute values of the first 100 vertices
        or edges. Numeric attributes will be aligned right, everything else will be
        aligned left."""
        values = []
        if vertex_attrs is not None:
            vs = self._graph.vs[:100]
            values.extend(vs[attr] for attr in vertex_attrs)
        if edge_attrs is not None:
            es = self._graph.es[:100]
            values.extend(es[attr] for attr in edge_attrs)

        result = []
        for vs in values:
            is_numeric = True
            try:
                [float(x) for x in vs]
            except ValueError:
                is_numeric = False
            if is_numeric:
                result.append("r")
            else:
                result.append("l")

        return result

    def _new_table(self, headers=None):
        """Constructs a new table to pretty-print vertex and edge attributes"""
        table = Texttable(max_width=0)
        table.set_deco(0)
        if headers is not None:
            table.header(headers)
        return table

    def _vertex_attribute_iterator(self, attribute_order):
        """Returns an iterator that yields the rows of the vertex attribute table
        in the summary. C{attribute_order} must be a list containing the names of
        the attributes to be presented in this table."""
        for vertex in self._graph.vs:
            yield ["[%d]" % vertex.index] + [vertex[attr] for attr in attribute_order]

    def __str__(self):
        """Returns the summary representation as a string."""
        output = self._construct_header()

        if self.print_graph_attributes:
            output.extend(self._construct_graph_attributes())
        if self.print_vertex_attributes:
            output.extend(self._construct_vertex_attributes())

        if self.verbosity <= 0:
            return "\n".join(output)

        if self._graph.ecount() > 0:
            # Add the edge list
            if self.edge_list_format == "auto":
                if self.print_edge_attributes and self._graph.edge_attributes():
                    format = "edgelist"
                elif median(self._graph.degree(mode="out")) < 3:
                    format = "compressed"
                else:
                    format = "adjlist"
            else:
                format = self.edge_list_format

            method_name = "_construct_edgelist_%s" % format
            if hasattr(self, method_name):
                output.extend(getattr(self, method_name)())

        if self.wrapper is not None:
            return "\n".join("\n".join(self.wrapper.wrap(line)) for line in output)

        return "\n".join(output)


def summary(obj, stream=None, *args, **kwds):
    """Prints a summary of object o to a given stream

    Positional and keyword arguments not explicitly mentioned here are passed
    on to the underlying C{summary()} method of the object if it has any.

    @param obj: the object about which a human-readable summary is requested.
    @param stream: the stream to be used. If C{None}, the standard output
      will be used.
    """
    if stream is None:
        stream = sys.stdout
    if hasattr(obj, "summary"):
        stream.write(obj.summary(*args, **kwds))
    else:
        stream.write(str(obj))
    stream.write("\n")
